import os

import pytest
from django.urls import reverse

from ...models import Theme


@pytest.fixture
def default_theme(db):
    return Theme.objects.get(is_default=True)


@pytest.fixture
def theme(db):
    return Theme.objects.create(name="Custom theme")


@pytest.fixture
def other_theme(db):
    return Theme.objects.create(name="Other theme")


@pytest.fixture
def nonexisting_theme(mocker, default_theme):
    return mocker.Mock(pk=default_theme.pk + 1)


TESTS_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def css(admin_client, theme, mock_build_theme_css):
    url = reverse("misago:admin:themes:upload-css", kwargs={"pk": theme.pk})
    with open(os.path.join(TESTS_DIR, "css", "test.css")) as fp:
        admin_client.post(url, {"assets": [fp]})
    return theme.css.get(name="test.css")


@pytest.fixture
def css_link(admin_client, theme):
    return theme.css.create(
        name="CSS link", url="https://example.com/cdn.css", order=theme.css.count()
    )


@pytest.fixture
def css_needing_build(admin_client, theme, mock_build_theme_css):
    url = reverse("misago:admin:themes:upload-css", kwargs={"pk": theme.pk})
    with open(os.path.join(TESTS_DIR, "css", "test.needs-build.css")) as fp:
        admin_client.post(url, {"assets": [fp]})
    return theme.css.get(name="test.needs-build.css")


@pytest.fixture
def media(admin_client, theme):
    url = reverse("misago:admin:themes:upload-media", kwargs={"pk": theme.pk})
    with open(os.path.join(TESTS_DIR, "images", "test.svg")) as fp:
        admin_client.post(url, {"assets": [fp]})
    return theme.media.get(name="test.svg")


@pytest.fixture
def image(admin_client, theme):
    url = reverse("misago:admin:themes:upload-media", kwargs={"pk": theme.pk})
    with open(os.path.join(TESTS_DIR, "images", "test.png"), "rb") as fp:
        admin_client.post(url, {"assets": [fp]})
    return theme.media.get(name="test.png")


@pytest.fixture(autouse=True)
def mock_build_single_theme_css(mocker):
    delay = mocker.Mock()
    mocker.patch(
        "misago.themes.admin.views.build_single_theme_css", mocker.Mock(delay=delay)
    )
    return delay


@pytest.fixture(autouse=True)
def mock_build_theme_css(mocker):
    delay = mocker.Mock()
    mocker.patch("misago.themes.admin.views.build_theme_css", mocker.Mock(delay=delay))
    mocker.patch(
        "misago.themes.admin.importer.build_theme_css", mocker.Mock(delay=delay)
    )
    return delay


@pytest.fixture(autouse=True)
def mock_update_remote_css_size(mocker):
    delay = mocker.Mock()
    mocker.patch(
        "misago.themes.admin.views.update_remote_css_size", mocker.Mock(delay=delay)
    )
    mocker.patch(
        "misago.themes.admin.importer.update_remote_css_size", mocker.Mock(delay=delay)
    )
    return delay



from django import forms
from django.utils.translation import gettext_lazy as _

from .base import ChangeSettingsForm


class ChangeThreadsSettingsForm(ChangeSettingsForm):
    settings = [
        "attachment_403_image",
        "attachment_404_image",
        "daily_post_limit",
        "hourly_post_limit",
        "post_attachments_limit",
        "post_length_max",
        "post_length_min",
        "readtracker_cutoff",
        "thread_title_length_max",
        "thread_title_length_min",
        "unused_attachments_lifetime",
        "threads_per_page",
        "posts_per_page",
        "posts_per_page_orphans",
        "events_per_page",
    ]

    daily_post_limit = forms.IntegerField(
        label=_("Daily post limit per user"),
        help_text=_(
            "Daily limit of posts that may be posted by single user. "
            "Fail-safe for situations when forum is flooded by spam bots. "
            "Change to 0 to remove the limit."
        ),
        min_value=0,
    )
    hourly_post_limit = forms.IntegerField(
        label=_("Hourly post limit per user"),
        help_text=_(
            "Hourly limit of posts that may be posted by single user. "
            "Fail-safe for situations when forum is flooded by spam bots. "
            "Change to 0 to remove the limit."
        ),
        min_value=0,
    )
    post_attachments_limit = forms.IntegerField(
        label=_("Maximum number of attachments per post"), min_value=1
    )
    post_length_max = forms.IntegerField(
        label=_("Maximum allowed post length"), min_value=0
    )
    post_length_min = forms.IntegerField(
        label=_("Minimum required post length"), min_value=1
    )
    thread_title_length_max = forms.IntegerField(
        label=_("Maximum allowed thread title length"), min_value=2, max_value=255
    )
    thread_title_length_min = forms.IntegerField(
        label=_("Minimum required thread title length"), min_value=2, max_value=255
    )
    unused_attachments_lifetime = forms.IntegerField(
        label=_("Unused attachments lifetime"),
        help_text=_(
            "Period of time (in hours) after which user-uploaded files that weren't "
            "attached to any post are deleted from disk."
        ),
        min_value=1,
    )

    readtracker_cutoff = forms.IntegerField(
        label=_("Read-tracker cutoff"),
        help_text=_(
            "Controls amount of data used by read-tracking system. All content older "
            "than number of days specified in this setting is considered old and read, "
            "even if the opposite is true for the user. Active forums can try lowering "
            "this value while less active ones may wish to increase it instead. "
        ),
        min_value=1,
    )

    threads_per_page = forms.IntegerField(
        label=_("Number of threads displayed on a single page"), min_value=10
    )

    posts_per_page = forms.IntegerField(
        label=_("Number of posts displayed on a single page"), min_value=5
    )
    posts_per_page_orphans = forms.IntegerField(
        label=_("Maximum orphans"),
        help_text=_(
            "If number of posts to be displayed on the last page is less or equal to "
            "number specified in this setting, those posts will instead be displayed "
            "on previous page, reducing the total number of pages in thread."
        ),
        min_value=0,
    )
    events_per_page = forms.IntegerField(
        label=_("Maximum number of events displayed on a single page"), min_value=5
    )

    attachment_403_image = forms.ImageField(
        label=_("Permission denied"),
        help_text=_(
            "Attachments proxy will display this image in place of default one "
            "when user tries to access attachment they have no permission to see."
        ),
        required=False,
    )
    attachment_403_image_delete = forms.BooleanField(
        label=_("Delete custom permission denied image"), required=False
    )
    attachment_404_image = forms.ImageField(
        label=_("Not found"),
        help_text=_(
            "Attachments proxy will display this image in place of default one "
            "when user tries to access attachment that doesn't exist."
        ),
        required=False,
    )
    attachment_404_image_delete = forms.BooleanField(
        label=_("Delete custom not found image"), required=False
    )

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("posts_per_page_orphans") > cleaned_data.get(
            "posts_per_page"
        ):
            self.add_error(
                "posts_per_page_orphans",
                _("This value must be lower than number of posts per page."),
            )
        return cleaned_data



from unittest.mock import Mock

import pytest

from ..middleware import cache_versions_middleware


@pytest.fixture
def get_response():
    return Mock()


@pytest.fixture
def request_mock():
    return Mock()


def test_middleware_sets_attr_on_request(db, get_response, request_mock):
    middleware = cache_versions_middleware(get_response)
    middleware(request_mock)
    assert request_mock.cache_versions


def test_middleware_calls_get_response(db, get_response, request_mock):
    middleware = cache_versions_middleware(get_response)
    middleware(request_mock)
    get_response.assert_called_once()



from django import forms
from django.utils.translation import gettext_lazy as _

from ....admin.forms import YesNoSwitch
from .base import ChangeSettingsForm


class ChangeGeneralSettingsForm(ChangeSettingsForm):
    settings = [
        "forum_name",
        "forum_address",
        "index_header",
        "index_title",
        "index_message",
        "index_meta_description",
        "logo",
        "logo_small",
        "logo_text",
        "og_image",
        "og_image_avatar_on_profile",
        "og_image_avatar_on_thread",
        "forum_footnote",
        "email_footer",
    ]

    forum_name = forms.CharField(label=_("Forum name"), min_length=2, max_length=255)
    forum_address = forms.URLField(label=_("Forum address"), max_length=255)

    index_title = forms.CharField(label=_("Page title"), max_length=255, required=False)
    index_meta_description = forms.CharField(
        label=_("Meta Description"),
        help_text=_(
            "Short description of your forum that search and social sites may "
            "display next to link to your forum's index."
        ),
        max_length=255,
        required=False,
    )
    index_header = forms.CharField(
        label=_("Header text"),
        help_text=_("This text will be displayed in page header on forum index."),
        max_length=255,
        required=False,
    )
    index_message = forms.CharField(
        label=_("Header message"),
        help_text=_(
            "This message will be displayed in page header on forum index, "
            "under the header text."
        ),
        max_length=2048,
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    logo = forms.ImageField(
        label=_("Logo"),
        help_text=_("Image that will displayed in forum navbar."),
        required=False,
    )
    logo_delete = forms.BooleanField(label=_("Delete logo image"), required=False)
    logo_small = forms.ImageField(
        label=_("Small logo"),
        help_text=_(
            "Image that will be displayed in compact forum navbar. "
            "When set, it will replace icon pointing to forum index."
        ),
        required=False,
    )
    logo_small_delete = forms.BooleanField(
        label=_("Delete small logo image"), required=False
    )
    logo_text = forms.CharField(
        label=_("Text"),
        help_text=_(
            "Text displayed in forum navbar. If logo image was uploaded, text will "
            "be displayed right next to it. Never displayed by the compact navbar."
        ),
        max_length=255,
        required=False,
    )

    og_image = forms.ImageField(
        label=_("Image"),
        help_text=_(
            "Custom image that will appear next to links to your forum posted on "
            "social sites. Facebook recommends that this image should be "
            "1200 pixels wide and 630 pixels tall."
        ),
        required=False,
    )
    og_image_delete = forms.BooleanField(label=_("Delete image"), required=False)
    og_image_avatar_on_profile = YesNoSwitch(
        label=_("Replace image with avatar on user profiles")
    )
    og_image_avatar_on_thread = YesNoSwitch(
        label=_("Replace image with avatar on threads")
    )

    forum_footnote = forms.CharField(
        label=_("Forum footnote"),
        help_text=_("Short message displayed in forum footer."),
        max_length=300,
        required=False,
    )
    email_footer = forms.CharField(
        label=_("E-mails footer"),
        help_text=_(
            "Optional short message included at the end of e-mails sent by forum."
        ),
        max_length=255,
        required=False,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        address = self.request.build_absolute_uri("/")
        self["forum_address"].help_text = _(
            "Misago uses this setting to build links in e-mails sent to site "
            'users. Address under which site is running appears to be "%(address)s".'
        ) % {"address": address}

    def clean_forum_address(self):
        return self.cleaned_data["forum_address"].lower()



from datetime import timedelta

from django.utils import timezone
from django.utils.translation import gettext as _

from . import PostingEndpoint, PostingInterrupt, PostingMiddleware

MIN_POSTING_INTERVAL = 3


class FloodProtectionMiddleware(PostingMiddleware):
    def use_this_middleware(self):
        return (
            not self.user_acl["can_omit_flood_protection"]
            and self.mode != PostingEndpoint.EDIT
        )

    def interrupt_posting(self, serializer):
        now = timezone.now()

        if self.user.last_posted_on:
            previous_post = now - self.user.last_posted_on
            if previous_post.total_seconds() < MIN_POSTING_INTERVAL:
                raise PostingInterrupt(
                    _("You can't post message so quickly after previous one.")
                )

        self.user.last_posted_on = timezone.now()
        self.user.update_fields.append("last_posted_on")

        if self.settings.hourly_post_limit:
            cutoff = now - timedelta(hours=1)
            if self.is_limit_exceeded(cutoff, self.settings.hourly_post_limit):
                raise PostingInterrupt(
                    _("Your account has exceed an hourly post limit.")
                )

        if self.settings.daily_post_limit:
            cutoff = now - timedelta(hours=24)
            if self.is_limit_exceeded(cutoff, self.settings.daily_post_limit):
                raise PostingInterrupt(_("Your account has exceed a daily post limit."))

    def is_limit_exceeded(self, cutoff, limit):
        return self.user.post_set.filter(posted_on__gte=cutoff).count() >= limit



from django.test import TestCase
from django.utils import timezone

from ...categories.models import Category
from ...users.test import create_test_user
from ..models import Post, Thread, ThreadParticipant


class ThreadParticipantTests(TestCase):
    def setUp(self):
        datetime = timezone.now()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.thread.set_title("Test thread")
        self.thread.save()

        post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster_name="Tester",
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.first_post = post
        self.thread.last_post = post
        self.thread.save()

    def test_set_owner(self):
        """set_owner makes user thread owner"""
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("User2", "user2@example.com")

        ThreadParticipant.objects.set_owner(self.thread, user)
        self.assertEqual(self.thread.participants.count(), 1)

        participant = ThreadParticipant.objects.get(thread=self.thread, user=user)
        self.assertTrue(participant.is_owner)
        self.assertEqual(user, participant.user)

        # threads can't have more than one owner
        ThreadParticipant.objects.set_owner(self.thread, other_user)
        self.assertEqual(self.thread.participants.count(), 2)

        participant = ThreadParticipant.objects.get(thread=self.thread, user=user)
        self.assertFalse(participant.is_owner)

        self.assertEqual(ThreadParticipant.objects.filter(is_owner=True).count(), 1)

    def test_add_participants(self):
        """add_participant adds participant to thread"""
        users = [
            create_test_user("User", "user@example.com"),
            create_test_user("User2", "user2@example.com"),
        ]

        ThreadParticipant.objects.add_participants(self.thread, users)
        self.assertEqual(self.thread.participants.count(), 2)

        for user in users:
            participant = ThreadParticipant.objects.get(thread=self.thread, user=user)
            self.assertFalse(participant.is_owner)

    def test_remove_participant(self):
        """remove_participant deletes participant from thread"""
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("User2", "user2@example.com")

        ThreadParticipant.objects.add_participants(self.thread, [user])
        ThreadParticipant.objects.add_participants(self.thread, [other_user])
        self.assertEqual(self.thread.participants.count(), 2)

        ThreadParticipant.objects.remove_participant(self.thread, user)
        self.assertEqual(self.thread.participants.count(), 1)

        with self.assertRaises(ThreadParticipant.DoesNotExist):
            ThreadParticipant.objects.get(thread=self.thread, user=user)



from django.contrib.auth import get_user_model
from django.urls import reverse

from ...acl.test import patch_user_acl
from ...admin.test import AdminTestCase

User = get_user_model()


class JoinIpProfileFieldTests(AdminTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:admin:users:edit", kwargs={"pk": self.user.pk})

    def test_field_hidden_in_admin(self):
        """readonly field doesn't display in the admin"""
        response = self.client.get(self.test_link)
        self.assertNotContains(response, 'name="join_ip"')
        self.assertNotContains(response, "IP address")
        self.assertNotContains(response, "Join IP")

    def test_admin_edits_field(self):
        """join_ip is non-editable by admin"""
        response = self.client.post(
            self.test_link,
            data={
                "username": "Edited",
                "rank": str(self.user.rank_id),
                "roles": str(self.user.roles.all()[0].pk),
                "email": "reg@stered.com",
                "join_ip": "127.0.0.1",
                "new_password": "",
                "signature": "",
                "is_signature_locked": "0",
                "is_hiding_presence": "0",
                "limits_private_thread_invites_to": "0",
                "signature_lock_staff_message": "",
                "signature_lock_user_message": "",
                "subscribe_to_started_threads": "2",
                "subscribe_to_replied_threads": "2",
            },
        )
        self.assertEqual(response.status_code, 302)

        self.reload_user()
        self.assertNotIn("join_ip", self.user.profile_fields)

    def test_admin_search_field(self):
        """admin users search searches this field"""
        test_link = reverse("misago:admin:users:index")

        response = self.client.get(
            "%s?redirected=1&profilefields=127.0.0.1" % test_link
        )
        self.assertContains(response, "No users matching criteria exist.")

    def test_field_display(self):
        """field displays on user profile"""
        test_link = reverse(
            "misago:user-details", kwargs={"pk": self.user.pk, "slug": self.user.slug}
        )

        response = self.client.get(test_link)
        self.assertContains(response, "IP address")
        self.assertContains(response, "Join IP")
        self.assertContains(response, "127.0.0.1")

    @patch_user_acl({"can_see_users_ips": 0})
    def test_field_hidden_no_permission(self):
        """field is hidden on user profile if user has no permission"""
        test_link = reverse(
            "misago:user-details", kwargs={"pk": self.user.pk, "slug": self.user.slug}
        )

        response = self.client.get(test_link)
        self.assertNotContains(response, "IP address")
        self.assertNotContains(response, "Join IP")
        self.assertNotContains(response, "127.0.0.1")

    def test_field_hidden_ip_removed(self):
        """field is hidden on user profile if ip is removed"""
        test_link = reverse(
            "misago:user-details", kwargs={"pk": self.user.pk, "slug": self.user.slug}
        )

        self.user.joined_from_ip = None
        self.user.save()

        response = self.client.get(test_link)
        self.assertNotContains(response, "IP address")
        self.assertNotContains(response, "Join IP")
        self.assertNotContains(response, "127.0.0.1")

    def test_field_display_json(self):
        """field is included in display json"""
        test_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)
        self.assertEqual(
            response.json()["groups"],
            [
                {
                    "name": "IP address",
                    "fields": [
                        {"fieldname": "join_ip", "name": "Join IP", "text": "127.0.0.1"}
                    ],
                }
            ],
        )

    @patch_user_acl({"can_see_users_ips": 0})
    def test_field_hidden_no_permission_json(self):
        """field is not included in display json if user has no permission"""
        test_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)
        self.assertEqual(response.json()["groups"], [])

    def test_field_hidden_ip_removed_json(self):
        """field is not included in display json if user ip is removed"""
        test_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk})

        self.user.joined_from_ip = None
        self.user.save()

        response = self.client.get(test_link)
        self.assertEqual(response.json()["groups"], [])

    def test_field_not_in_edit_json(self):
        """readonly field json is not returned from API"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)

        found_field = None
        for group in response.json():
            for field in group["fields"]:
                if field["fieldname"] == "join_ip":
                    found_field = field

        self.assertIsNone(found_field)

    def test_field_is_not_editable_in_api(self):
        """readonly field can't be edited via api"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        response = self.client.post(test_link, data={"join_ip": "88.12.13.14"})
        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertNotIn("join_ip", self.user.profile_fields)



from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.response import Response

from ....core.mail import mail_user
from ...credentialchange import store_new_credential
from ...serializers import ChangeEmailSerializer


def change_email_endpoint(request, pk=None):
    serializer = ChangeEmailSerializer(
        data=request.data, context={"user": request.user}
    )

    if not serializer.is_valid():
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    token = store_new_credential(
        request, "email", serializer.validated_data["new_email"]
    )

    mail_subject = _("Confirm e-mail change on %(forum_name)s forums")
    mail_subject = mail_subject % {"forum_name": request.settings.forum_name}

    # swap address with new one so email is sent to new address
    request.user.email = serializer.validated_data["new_email"]

    mail_user(
        request.user,
        mail_subject,
        "misago/emails/change_email",
        context={"settings": request.settings, "token": token},
    )

    message = _("E-mail change confirmation link was sent to new address.")
    return Response({"detail": message})



from django import forms
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from ...admin.forms import YesNoSwitch
from ..models import Agreement
from .utils import disable_agreement, set_agreement_as_active


class AgreementForm(forms.ModelForm):
    type = forms.ChoiceField(label=_("Type"), choices=Agreement.TYPE_CHOICES)
    title = forms.CharField(
        label=_("Title"),
        help_text=_("Optional, leave empty for agreement to be named after its type."),
        required=False,
    )
    is_active = YesNoSwitch(
        label=_("Active for its type"),
        help_text=_(
            "If other agreement is already active for this type, it will be unset "
            "and replaced with this one. "
            "Misago will ask users who didn't accept this agreement to do so "
            "before allowing them to continue using the site."
        ),
    )
    link = forms.URLField(
        label=_("Link"),
        help_text=_(
            "If your agreement is located on other page, enter here a link to it."
        ),
        required=False,
    )
    text = forms.CharField(
        label=_("Text"),
        help_text=_("You can use Markdown syntax for rich text elements."),
        widget=forms.Textarea,
        required=False,
    )

    class Meta:
        model = Agreement
        fields = ["type", "title", "link", "text", "is_active"]

    def clean(self):
        data = super().clean()

        if not data.get("link") and not data.get("text"):
            raise forms.ValidationError(_("Please fill in agreement link or text."))

        return data

    def save(self):
        agreement = super().save()
        if agreement.is_active:
            set_agreement_as_active(agreement)
        else:
            disable_agreement(agreement)
        Agreement.objects.invalidate_cache()
        return agreement


class FilterAgreementsForm(forms.Form):
    type = forms.ChoiceField(
        label=_("Type"),
        required=False,
        choices=[("", _("All types"))] + Agreement.TYPE_CHOICES,
    )
    content = forms.CharField(label=_("Content"), required=False)

    def filter_queryset(self, criteria, queryset):
        if criteria.get("type") is not None:
            queryset = queryset.filter(type=criteria["type"])

        if criteria.get("content"):
            search_title = Q(title__icontains=criteria["content"])
            search_text = Q(text__icontains=criteria["content"])
            queryset = queryset.filter(search_title | search_text)

        return queryset



"""
Supported inline BBCodes: b, u, i
"""
import re
from xml.etree.ElementTree import Element

from markdown.inlinepatterns import (
    ImageInlineProcessor,
    LinkInlineProcessor,
    Pattern,
    SimpleTagPattern,
    dequote,
)


class SimpleBBCodePattern(SimpleTagPattern):
    """
    Case insensitive simple BBCode
    """

    def __init__(self, bbcode, tag=None):  # pylint: disable=super-init-not-called
        self.pattern = r"(\[%s\](.*?)\[/%s\])" % (bbcode, bbcode)
        self.compiled_re = re.compile(
            "^(.*?)%s(.*?)$" % self.pattern, re.DOTALL | re.UNICODE | re.IGNORECASE
        )

        # Api for Markdown to pass safe_mode into instance
        self.safe_mode = False

        # Store tag
        self.tag = tag or bbcode.lower()


bold = SimpleBBCodePattern("b")
italics = SimpleBBCodePattern("i")
underline = SimpleBBCodePattern("u")


class BBcodeProcessor(Pattern):
    def __init__(self, pattern, md=None):
        self.pattern = pattern
        self.compiled_re = re.compile(
            "^(.*?)%s(.*)$" % pattern, re.DOTALL | re.UNICODE | re.IGNORECASE
        )

        self.safe_mode = False
        if md:
            self.md = md


class BBCodeImageProcessor(BBcodeProcessor, ImageInlineProcessor):
    def handleMatch(self, m, _):
        el = Element("img")

        src = m.group("content").strip()
        el.set("src", self.unescape(src))

        alt_text = src.replace('"', "&quot;")
        if alt_text.lower()[:6] == "https:":
            alt_text = alt_text[6:]
        elif alt_text.lower()[:5] == "http:":
            alt_text = alt_text[5:]

        alt_text = alt_text.lstrip("/")

        el.set("alt", alt_text)

        return el, m.start("open"), m.end("close")


IMAGE_PATTERN = r"(?P<open>\[img\])(?P<content>.*?)(?P<close>\[/img\])"


def image(md):
    return BBCodeImageProcessor(IMAGE_PATTERN, md)


class BBCodeUrlPattern(BBcodeProcessor, LinkInlineProcessor):
    def handleMatch(self, m, _):
        el = Element("a")

        if m.group("arg"):
            el.text = m.group("content")
            href = m.group("arg")
        else:
            el.text = m.group("content").strip()
            href = m.group("content")

        if href:
            el.set("href", self.unescape(href.strip()))
        else:
            el.set("href", "")

        return el, m.start("open"), m.end("close")


URL_PATTERN = r'(?P<open>(\[url=("?)(?P<arg>.*?)("?)\])|(\[url\]))(?P<content>.*?)(?P<close>\[/url\])'


def url(md):
    return BBCodeUrlPattern(URL_PATTERN, md)



def chunk_queryset(queryset, chunk_size=20):
    ordered_queryset = queryset.order_by("-pk")  # bias to newest items first
    chunk = ordered_queryset[:chunk_size]
    while chunk:
        last_pk = None
        for item in chunk:
            last_pk = item.pk
            yield item
        chunk = ordered_queryset.filter(pk__lt=last_pk)[:chunk_size]



import pytest

from django.core.files.uploadedfile import SimpleUploadedFile
from django.urls import reverse

from ...models import Setting

admin_link = reverse("misago:admin:settings:analytics:index")


@pytest.fixture
def setting(db):
    return Setting.objects.get(setting="google_site_verification")


@pytest.fixture
def setting_with_value(setting):
    setting.dry_value = "asdfghjkl1234567"
    setting.save()
    return setting


def test_site_verification_cant_be_edited_directly(admin_client, setting):
    admin_client.post(admin_link, {"google_site_verification": "test"})
    setting.refresh_from_db()
    assert not setting.value


def test_site_verification_is_set_from_uploaded_file(admin_client, setting):
    verification = b"google-site-verification: googleasdfghjkl1234567.html"
    verification_file = SimpleUploadedFile("test.html", verification, "text/html")
    admin_client.post(admin_link, {"google_site_verification_file": verification_file})

    setting.refresh_from_db()
    assert setting.value == "asdfghjkl1234567"


def test_non_html_uploaded_file_is_rejected(admin_client, setting):
    verification_file = SimpleUploadedFile("test.html", b"test", "text/plain")
    admin_client.post(admin_link, {"google_site_verification_file": verification_file})

    setting.refresh_from_db()
    assert not setting.value


def test_empty_uploaded_file_is_rejected(admin_client, setting_with_value):
    verification_file = SimpleUploadedFile("test.html", b"", "text/html")
    admin_client.post(admin_link, {"google_site_verification_file": verification_file})

    setting_with_value.refresh_from_db()
    assert setting_with_value.value


def test_incorrect_uploaded_file_is_rejected(admin_client, setting_with_value):
    verification = b"google-site-verification: google.html"
    verification_file = SimpleUploadedFile("test.html", verification, "text/html")
    admin_client.post(admin_link, {"google_site_verification_file": verification_file})

    setting_with_value.refresh_from_db()
    assert setting_with_value.value



from django.urls import reverse

from ....acl import ACL_CACHE
from ....acl.models import Role
from ....admin.test import AdminTestCase
from ....cache.test import assert_invalidates_cache
from ...models import Rank


class RankAdminTests(AdminTestCase):
    def test_link_registered(self):
        """admin nav contains ranks link"""
        response = self.client.get(reverse("misago:admin:users:index"))

        response = self.client.get(response["location"])
        self.assertContains(response, reverse("misago:admin:ranks:index"))

    def test_list_view(self):
        """ranks list view returns 200"""
        response = self.client.get(reverse("misago:admin:ranks:index"))

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Team")

    def test_new_view(self):
        """new rank view has no showstoppers"""
        test_role_a = Role.objects.create(name="Test Role A")
        test_role_b = Role.objects.create(name="Test Role B")
        test_role_c = Role.objects.create(name="Test Role C")

        response = self.client.get(reverse("misago:admin:ranks:new"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
                "roles": [test_role_a.pk, test_role_c.pk],
            },
        )
        self.assertEqual(response.status_code, 302)

        response = self.client.get(reverse("misago:admin:ranks:index"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Test Rank")
        self.assertContains(response, "Test Title")

        test_rank = Rank.objects.get(slug="test-rank")
        self.assertIn(test_role_a, test_rank.roles.all())
        self.assertIn(test_role_c, test_rank.roles.all())
        self.assertTrue(test_role_b not in test_rank.roles.all())

    def test_edit_view(self):
        """edit rank view has no showstoppers"""
        test_role_a = Role.objects.create(name="Test Role A")
        test_role_b = Role.objects.create(name="Test Role B")
        test_role_c = Role.objects.create(name="Test Role C")

        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
                "roles": [test_role_a.pk, test_role_c.pk],
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.get(
            reverse("misago:admin:ranks:edit", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, test_rank.name)
        self.assertContains(response, test_rank.title)

        response = self.client.post(
            reverse("misago:admin:ranks:edit", kwargs={"pk": test_rank.pk}),
            data={"name": "Top Lel", "roles": [test_role_b.pk]},
        )
        self.assertEqual(response.status_code, 302)

        test_rank = Rank.objects.get(slug="top-lel")
        response = self.client.get(reverse("misago:admin:ranks:index"))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, test_rank.name)
        self.assertTrue("Test Rank" not in test_rank.roles.all())
        self.assertTrue("Test Title" not in test_rank.roles.all())

        self.assertIn(test_role_b, test_rank.roles.all())
        self.assertTrue(test_role_a not in test_rank.roles.all())
        self.assertTrue(test_role_c not in test_rank.roles.all())

    def test_editing_rank_invalidates_acl_cache(self):
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")
        test_role_b = Role.objects.create(name="Test Role B")

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse("misago:admin:ranks:edit", kwargs={"pk": test_rank.pk}),
                data={"name": "Top Lel", "roles": [test_role_b.pk]},
            )

    def test_default_view(self):
        """default rank view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.post(
            reverse("misago:admin:ranks:default", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 302)

        test_rank = Rank.objects.get(slug="test-rank")
        self.assertTrue(test_rank.is_default)

    def test_move_up_view(self):
        """move rank up view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.post(
            reverse("misago:admin:ranks:up", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 302)

        changed_rank = Rank.objects.get(slug="test-rank")
        self.assertEqual(changed_rank.order + 1, test_rank.order)

    def test_move_down_view(self):
        """move rank down view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        # Move rank up
        response = self.client.post(
            reverse("misago:admin:ranks:up", kwargs={"pk": test_rank.pk})
        )

        response = self.client.post(
            reverse("misago:admin:ranks:down", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 302)

        # Test move down
        changed_rank = Rank.objects.get(slug="test-rank")
        self.assertEqual(changed_rank.order, test_rank.order)

    def test_users_view(self):
        """users with this rank view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.get(
            reverse("misago:admin:ranks:users", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 302)

    def test_delete_view(self):
        """delete rank view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.post(
            reverse("misago:admin:ranks:delete", kwargs={"pk": test_rank.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.client.get(reverse("misago:admin:ranks:index"))
        response = self.client.get(reverse("misago:admin:ranks:index"))
        self.assertEqual(response.status_code, 200)

        self.assertNotContains(response, test_rank.name)
        self.assertNotContains(response, test_rank.title)

    def test_deleting_rank_invalidates_acl_cache(self):
        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test Rank",
                "description": "Lorem ipsum dolor met",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse("misago:admin:ranks:delete", kwargs={"pk": test_rank.pk})
            )

    def test_uniquess(self):
        """rank slug uniqueness is enforced by admin forms"""
        test_role_a = Role.objects.create(name="Test Role A")

        response = self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Members",
                "description": "Colliding rank",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
                "roles": [test_role_a.pk],
            },
        )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "This name collides with other rank.")

        self.client.post(
            reverse("misago:admin:ranks:new"),
            data={
                "name": "Test rank",
                "description": "Colliding rank",
                "title": "Test Title",
                "style": "test",
                "is_tab": "1",
                "roles": [test_role_a.pk],
            },
        )

        test_rank = Rank.objects.get(slug="test-rank")

        response = self.client.post(
            reverse("misago:admin:ranks:edit", kwargs={"pk": test_rank.pk}),
            data={"name": "Members", "roles": [test_role_a.pk]},
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "This name collides with other rank.")



from django.template import Context, Template
from django.test import TestCase

from ..templatetags.misago_batch import batch, batchnonefilled


class CaptureTests(TestCase):
    def setUp(self):
        self.context = Context({"unsafe_name": "The<hr>Html"})

    def test_capture(self):
        """capture content to variable"""
        tpl_content = """
{% load misago_capture %}

{% capture as the_var %}
{{ unsafe_name }}
{% endcapture %}
Hello, <b>{{ the_var|safe }}</b>
"""

        tpl = Template(tpl_content)
        render = tpl.render(self.context).strip()
        self.assertIn("The&lt;hr&gt;Html", render)
        self.assertNotIn("<b>The&lt;hr&gt;Html</b>", render)

    def test_capture_trimmed(self):
        """capture trimmed content to variable"""
        tpl_content = """
{% load misago_capture %}

{% capture trimmed as the_var %}
{{ unsafe_name }}
{% endcapture %}
Hello, <b>{{ the_var|safe }}</b>
"""

        tpl = Template(tpl_content)
        render = tpl.render(self.context).strip()
        self.assertIn("<b>The&lt;hr&gt;Html</b>", render)


class BatchTests(TestCase):
    def test_batch(self):
        """standard batch yields valid results"""
        value = "loremipsum"
        result = [["l", "o", "r"], ["e", "m", "i"], ["p", "s", "u"], ["m"]]

        for i, test_result in enumerate(batch(value, 3)):
            self.assertEqual(test_result, result[i])

    def test_batchnonefilled(self):
        """none-filled batch yields valid results"""
        value = "loremipsum"
        result = [["l", "o", "r"], ["e", "m", "i"], ["p", "s", "u"], ["m", None, None]]

        for i, test_result in enumerate(batchnonefilled(value, 3)):
            self.assertEqual(test_result, result[i])


class MockUser:
    id = 12
    pk = 12
    username = "User"
    slug = "user"


class ShorthandsTests(TestCase):
    def test_iftrue_for_true(self):
        """iftrue renders value for true"""
        tpl_content = """
{% load misago_shorthands %}

{{ value|iftrue:result }}
"""

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"result": "Ok!", "value": True})).strip(), "Ok!"
        )

    def test_iftrue_for_false(self):
        """iftrue isnt rendering value for false"""
        tpl_content = """
{% load misago_shorthands %}

{{ value|iftrue:result }}
"""

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"result": "Ok!", "value": False})).strip(), ""
        )

    def test_iffalse_for_true(self):
        """iffalse isnt rendering value for true"""
        tpl_content = """
{% load misago_shorthands %}

{{ value|iffalse:result }}
"""

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"result": "Ok!", "value": True})).strip(), ""
        )

    def test_iffalse_for_false(self):
        """iffalse renders value for false"""
        tpl_content = """
{% load misago_shorthands %}

{{ value|iffalse:result }}
"""

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"result": "Ok!", "value": False})).strip(), "Ok!"
        )


class JSONTests(TestCase):
    def test_json_filter(self):
        """as_json filter renders dict as safe json"""
        tpl_content = """
{% load misago_json %}

{{ value|as_json }}
"""

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"value": {"he</script>llo": 'bo"b!'}})).strip(),
            r'{"he\u003C/script>llo": "bo\"b!"}',
        )


class PageTitleTests(TestCase):
    def test_single_title(self):
        """tag passes trough single title"""
        tpl_content = """
        {% load misago_pagetitle %}

        {% pagetitle item %}
        """

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"item": "Lorem Ipsum"})).strip(), "Lorem Ipsum"
        )

    def test_parent_title(self):
        """tag builds full title from title and parent name"""
        tpl_content = """
        {% load misago_pagetitle %}

        {% pagetitle item parent=parent %}
        """

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(
                Context({"item": "Lorem Ipsum", "parent": "Some Thread"})
            ).strip(),
            "Lorem Ipsum | Some Thread",
        )

    def test_paged_title(self):
        """tag builds full title from title and page number"""
        tpl_content = """
        {% load misago_pagetitle %}

        {% pagetitle item page=3 %}
        """

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(Context({"item": "Lorem Ipsum"})).strip(),
            "Lorem Ipsum (page: 3)",
        )

    def test_kitchensink_title(self):
        """tag builds full title from all options"""
        tpl_content = """
        {% load misago_pagetitle %}

        {% pagetitle item page=3 parent=parent %}
        """

        tpl = Template(tpl_content)
        self.assertEqual(
            tpl.render(
                Context({"item": "Lorem Ipsum", "parent": "Some Thread"})
            ).strip(),
            "Lorem Ipsum (page: 3) | Some Thread",
        )



from django.urls import reverse

from ...acl.test import patch_user_acl
from ..models import Poll, Thread
from ..serializers.poll import MAX_POLL_OPTIONS
from ..test import patch_category_acl
from .test_thread_poll_api import ThreadPollApiTestCase


class ThreadPollCreateTests(ThreadPollApiTestCase):
    def test_anonymous(self):
        """api requires you to sign in to create poll"""
        self.logout_user()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)

    def test_invalid_thread_id(self):
        """api validates that thread id is integer"""
        api_link = reverse(
            "misago:api:thread-poll-list", kwargs={"thread_pk": "kjha6dsa687sa"}
        )

        response = self.post(api_link)
        self.assertEqual(response.status_code, 404)

    def test_nonexistant_thread_id(self):
        """api validates that thread exists"""
        api_link = reverse(
            "misago:api:thread-poll-list", kwargs={"thread_pk": self.thread.pk + 1}
        )

        response = self.post(api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_start_polls": 0})
    def test_no_permission(self):
        """api validates that user has permission to start poll in thread"""
        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't start polls."})

    @patch_user_acl({"can_start_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_closed_thread_no_permission(self):
        """api validates that user has permission to start poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't start polls in it."},
        )

    @patch_user_acl({"can_start_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_thread(self):
        """api validates that user has permission to start poll in closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_user_acl({"can_start_polls": 1})
    @patch_category_acl({"can_close_threads": False})
    def test_closed_category_no_permission(self):
        """api validates that user has permission to start poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't start polls in it."},
        )

    @patch_user_acl({"can_start_polls": 1})
    @patch_category_acl({"can_close_threads": True})
    def test_closed_category(self):
        """api validates that user has permission to start poll in closed category"""
        self.category.is_closed = True
        self.category.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    @patch_user_acl({"can_start_polls": 1})
    def test_other_user_thread_no_permission(self):
        """
        api validates that user has permission to start poll in other user's thread
        """
        self.thread.starter = None
        self.thread.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't start polls in other users threads."}
        )

    @patch_user_acl({"can_start_polls": 2})
    def test_other_user_thread(self):
        """
        api validates that user has permission to start poll in other user's thread
        """
        self.thread.starter = None
        self.thread.save()

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

    def test_no_permission_poll_exists(self):
        """api validates that user can't start second poll in thread"""
        self.thread.poll = Poll.objects.create(
            thread=self.thread,
            category=self.category,
            poster_name="Test",
            poster_slug="test",
            length=30,
            question="Test",
            choices=[{"hash": "t3st"}],
            allowed_choices=1,
        )

        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "There's already a poll in this thread."}
        )

    def test_empty_data(self):
        """api handles empty request data"""
        response = self.post(self.api_link)
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(len(response_json), 4)

    def test_length_validation(self):
        """api validates poll's length"""
        response = self.post(self.api_link, data={"length": -1})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["length"],
            ["Ensure this value is greater than or equal to 0."],
        )

        response = self.post(self.api_link, data={"length": 200})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["length"], ["Ensure this value is less than or equal to 180."]
        )

    def test_question_validation(self):
        """api validates question length"""
        response = self.post(self.api_link, data={"question": "abcd" * 255})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["question"],
            ["Ensure this field has no more than 255 characters."],
        )

    def test_validate_choice_length(self):
        """api validates single choice length"""
        response = self.post(
            self.api_link, data={"choices": [{"hash": "qwertyuiopas", "label": ""}]}
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"], ["One or more poll choices are invalid."]
        )

        response = self.post(
            self.api_link,
            data={"choices": [{"hash": "qwertyuiopas", "label": "abcd" * 255}]},
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"], ["One or more poll choices are invalid."]
        )

    def test_validate_two_choices(self):
        """api validates that there are at least two choices in poll"""
        response = self.post(self.api_link, data={"choices": [{"label": "Choice"}]})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"],
            ["You need to add at least two choices to a poll."],
        )

    def test_validate_max_choices(self):
        """api validates that there are no more choices in poll than allowed number"""
        response = self.post(
            self.api_link,
            data={"choices": [{"label": "Choice"}] * (MAX_POLL_OPTIONS + 1)},
        )
        self.assertEqual(response.status_code, 400)

        error_formats = (MAX_POLL_OPTIONS, MAX_POLL_OPTIONS + 1)

        response_json = response.json()
        self.assertEqual(
            response_json["choices"],
            [
                "You can't add more than %s options to a single poll (added %s)."
                % error_formats
            ],
        )

    def test_allowed_choices_validation(self):
        """api validates allowed choices number"""
        response = self.post(self.api_link, data={"allowed_choices": 0})
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["allowed_choices"],
            ["Ensure this value is greater than or equal to 1."],
        )

        response = self.post(
            self.api_link,
            data={
                "length": 0,
                "question": "Lorem ipsum",
                "allowed_choices": 3,
                "choices": [{"label": "Choice"}, {"label": "Choice"}],
            },
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json["non_field_errors"],
            ["Number of allowed choices can't be greater than number of all choices."],
        )

    def test_poll_created(self):
        """api creates public poll if provided with valid data"""
        response = self.post(
            self.api_link,
            data={
                "length": 40,
                "question": "Select two best colors",
                "allowed_choices": 2,
                "allow_revotes": True,
                "is_public": True,
                "choices": [{"label": "\nRed "}, {"label": "Green"}, {"label": "Blue"}],
            },
        )
        self.assertEqual(response.status_code, 200)

        response_json = response.json()

        self.assertEqual(response_json["poster_name"], self.user.username)
        self.assertEqual(response_json["length"], 40)
        self.assertEqual(response_json["question"], "Select two best colors")
        self.assertEqual(response_json["allowed_choices"], 2)
        self.assertTrue(response_json["allow_revotes"])
        self.assertEqual(response_json["votes"], 0)
        self.assertTrue(response_json["is_public"])

        self.assertEqual(len(response_json["choices"]), 3)
        self.assertEqual(len({c["hash"] for c in response_json["choices"]}), 3)
        self.assertEqual(
            [c["label"] for c in response_json["choices"]], ["Red", "Green", "Blue"]
        )

        thread = Thread.objects.get(pk=self.thread.pk)
        self.assertTrue(thread.has_poll)

        poll = thread.poll

        self.assertEqual(poll.category_id, self.category.id)
        self.assertEqual(poll.thread_id, self.thread.id)
        self.assertEqual(poll.poster_id, self.user.id)
        self.assertEqual(poll.poster_name, self.user.username)
        self.assertEqual(poll.poster_slug, self.user.slug)
        self.assertEqual(poll.length, 40)
        self.assertEqual(poll.question, "Select two best colors")
        self.assertEqual(poll.allowed_choices, 2)
        self.assertTrue(poll.allow_revotes)
        self.assertEqual(poll.votes, 0)
        self.assertTrue(poll.is_public)

        self.assertEqual(len(poll.choices), 3)
        self.assertEqual(len({c["hash"] for c in poll.choices}), 3)

        self.assertEqual(self.user.audittrail_set.count(), 1)



from importlib import import_module

from .. import hooks
from ..conf import settings
from .htmlparser import parse_html_string, print_html_string


class MarkupPipeline:
    """small framework for extending parser"""

    def extend_markdown(self, md):
        for extension in settings.MISAGO_MARKUP_EXTENSIONS:
            module = import_module(extension)
            if hasattr(module, "extend_markdown"):
                hook = getattr(module, "extend_markdown")
                hook.extend_markdown(md)

        for extension in hooks.markdown_extensions:
            extension(md)

        return md

    def process_result(self, result):
        if (
            not settings.MISAGO_MARKUP_EXTENSIONS
            and not hooks.parsing_result_processors
        ):
            return result

        html_tree = parse_html_string(result["parsed_text"])
        for extension in settings.MISAGO_MARKUP_EXTENSIONS:
            module = import_module(extension)
            if hasattr(module, "clean_parsed"):
                hook = getattr(module, "clean_parsed")
                hook.process_result(result, html_tree)

        for extension in hooks.parsing_result_processors:
            extension(result, html_tree)

        result["parsed_text"] = print_html_string(html_tree)
        return result


pipeline = MarkupPipeline()



from unittest.mock import Mock, patch

import pytest
from django.contrib.auth import get_user_model

from ..exceptions import OAuth2UserDataValidationError
from ..models import Subject
from ..user import get_user_from_data

User = get_user_model()


def test_user_is_updated_with_valid_data(user, dynamic_settings):
    Subject.objects.create(sub="1234", user=user)

    updated_user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": "UpdatedName",
            "email": "updated@example.com",
            "avatar": None,
        },
        {},
    )

    assert created is False
    assert updated_user.id
    assert updated_user.id == user.id
    assert updated_user.username == "UpdatedName"
    assert updated_user.username != user.username
    assert updated_user.slug == "updatedname"
    assert updated_user.slug != user.slug
    assert updated_user.email == "updated@example.com"
    assert updated_user.email != user.email

    user_by_name = User.objects.get_by_username("UpdatedName")
    assert user_by_name.id == user.id

    user_by_email = User.objects.get_by_email("updated@example.com")
    assert user_by_email.id == user.id


def test_user_is_not_updated_with_unchanged_valid_data(user, dynamic_settings):
    Subject.objects.create(sub="1234", user=user)

    updated_user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": user.username,
            "email": user.email,
            "avatar": None,
        },
        {},
    )

    assert created is False
    assert updated_user.id
    assert updated_user.id == user.id
    assert updated_user.username == "User"
    assert updated_user.username == user.username
    assert updated_user.slug == "user"
    assert updated_user.slug == user.slug
    assert updated_user.email == "user@example.com"
    assert updated_user.email == user.email

    user_by_name = User.objects.get_by_username("User")
    assert user_by_name.id == user.id

    user_by_email = User.objects.get_by_email("user@example.com")
    assert user_by_email.id == user.id


def user_noop_filter(*args):
    pass


def test_user_name_conflict_during_update_with_valid_data_is_handled(
    user, other_user, dynamic_settings
):
    Subject.objects.create(sub="1234", user=user)

    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        # Custom filters disable build in filters
        with patch(
            "misago.oauth2.validation.oauth2_user_data_filters",
            [user_noop_filter],
        ):
            get_user_from_data(
                Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
                {
                    "id": "1234",
                    "name": other_user.username,
                    "email": "test@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == ["This username is not available."]


def test_user_email_conflict_during_update_with_valid_data_is_handled(
    user, other_user, dynamic_settings
):
    Subject.objects.create(sub="1234", user=user)

    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        get_user_from_data(
            Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
            {
                "id": "1234",
                "name": "NewUser",
                "email": other_user.email,
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["This e-mail address is not available."]



import json

from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.response import Response

from ... import avatars
from ....conf import settings
from ....core.decorators import require_dict_data
from ....core.utils import format_plaintext_for_html
from ...models import AvatarGallery
from ...serializers import ModerateAvatarSerializer


@require_dict_data
def avatar_endpoint(request, pk=None):
    if request.user.is_avatar_locked:
        if request.user.avatar_lock_user_message:
            reason = format_plaintext_for_html(request.user.avatar_lock_user_message)
        else:
            reason = None

        return Response(
            {
                "detail": _("Your avatar is locked. You can't change it."),
                "reason": reason,
            },
            status=status.HTTP_403_FORBIDDEN,
        )

    avatar_options = get_avatar_options(request, request.user)
    if request.method == "POST":
        return avatar_post(request, avatar_options)
    return Response(avatar_options)


def get_avatar_options(request, user):
    options = {
        "avatars": user.avatars,
        "generated": True,
        "gravatar": False,
        "crop_src": False,
        "crop_tmp": False,
        "upload": False,
        "galleries": False,
    }

    # Allow existing galleries
    if avatars.gallery.galleries_exist():
        options["galleries"] = []
        for gallery in avatars.gallery.get_available_galleries():
            gallery_images = []
            for image in gallery["images"]:
                gallery_images.append({"id": image.id, "url": image.url})
            options["galleries"].append(
                {"name": gallery["name"], "images": gallery_images}
            )

    # Can't have custom avatar?
    if not request.settings.allow_custom_avatars:
        return options

    # Allow Gravatar download
    options["gravatar"] = True

    # Allow crop if we have uploaded temporary avatar
    if avatars.uploaded.has_source_avatar(user):
        try:
            options["crop_src"] = {
                "url": user.avatar_src.url,
                "crop": json.loads(user.avatar_crop),
                "size": max(settings.MISAGO_AVATARS_SIZES),
            }
        except (TypeError, ValueError):
            pass

    # Allow crop of uploaded avatar
    if avatars.uploaded.has_temporary_avatar(user):
        options["crop_tmp"] = {
            "url": user.avatar_tmp.url,
            "size": max(settings.MISAGO_AVATARS_SIZES),
        }

    # Allow upload conditions
    options["upload"] = {
        "limit": request.settings.avatar_upload_limit * 1024,
        "allowed_extensions": avatars.uploaded.ALLOWED_EXTENSIONS,
        "allowed_mime_types": avatars.uploaded.ALLOWED_MIME_TYPES,
    }

    return options


class AvatarError(Exception):
    pass


def avatar_post(request, options):
    user = request.user
    data = request.data

    avatar_type = data.get("avatar", "nope")

    try:
        type_options = options[avatar_type]
        if not type_options:
            return Response(
                {"detail": _("This avatar type is not allowed.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        avatar_strategy = AVATAR_TYPES[avatar_type]
    except KeyError:
        return Response(
            {"detail": _("Unknown avatar type.")}, status=status.HTTP_400_BAD_REQUEST
        )

    try:
        if avatar_type == "upload":
            # avatar_upload strategy requires access to request.settings
            response_dict = {"detail": avatar_upload(request, user, data)}
        else:
            response_dict = {"detail": avatar_strategy(user, data)}
    except AvatarError as e:
        return Response({"detail": e.args[0]}, status=status.HTTP_400_BAD_REQUEST)

    user.save()

    updated_options = get_avatar_options(request, user)
    response_dict.update(updated_options)
    return Response(response_dict)


def avatar_generate(user, data):
    avatars.dynamic.set_avatar(user)
    return _("New avatar based on your account was set.")


def avatar_gravatar(user, data):
    try:
        avatars.gravatar.set_avatar(user)
        return _("Gravatar was downloaded and set as new avatar.")
    except avatars.gravatar.NoGravatarAvailable:
        raise AvatarError(_("No Gravatar is associated with your e-mail address."))
    except avatars.gravatar.GravatarError:
        raise AvatarError(_("Failed to connect to Gravatar servers."))


def avatar_gallery(user, data):
    try:
        image_pk = int(data.get("image"))
        image = AvatarGallery.objects.get(pk=image_pk)
        if image.gallery == "__default__":
            raise ValueError()
        avatars.gallery.set_avatar(user, image)
        return _("Avatar from gallery was set.")
    except (TypeError, ValueError, AvatarGallery.DoesNotExist):
        raise AvatarError(_("Incorrect image."))


def avatar_upload(request, user, data):
    new_avatar = data.get("image")
    if not new_avatar:
        raise AvatarError(_("No file was sent."))

    try:
        avatars.uploaded.handle_uploaded_file(request, user, new_avatar)
    except ValidationError as e:
        raise AvatarError(e.args[0])

    # send back url for temp image
    return user.avatar_tmp.url


def avatar_crop_src(user, data):
    avatar_crop(user, data, "src")
    return _("Avatar was re-cropped.")


def avatar_crop_tmp(user, data):
    avatar_crop(user, data, "tmp")
    return _("Uploaded avatar was set.")


def avatar_crop(user, data, suffix):
    try:
        crop = avatars.uploaded.crop_source_image(user, suffix, data.get("crop", {}))
        user.avatar_crop = json.dumps(crop)
    except ValidationError as e:
        raise AvatarError(e.args[0])


AVATAR_TYPES = {
    "generated": avatar_generate,
    "gravatar": avatar_gravatar,
    "galleries": avatar_gallery,
    "upload": avatar_upload,
    "crop_src": avatar_crop_src,
    "crop_tmp": avatar_crop_tmp,
}


def moderate_avatar_endpoint(request, profile):
    if request.method == "POST":
        is_avatar_locked = profile.is_avatar_locked
        serializer = ModerateAvatarSerializer(profile, data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        if serializer.validated_data["is_avatar_locked"] and not is_avatar_locked:
            avatars.dynamic.set_avatar(profile)
        serializer.save()

        return Response(
            {
                "avatars": profile.avatars,
                "is_avatar_locked": int(profile.is_avatar_locked),
                "avatar_lock_user_message": profile.avatar_lock_user_message,
                "avatar_lock_staff_message": profile.avatar_lock_staff_message,
            }
        )

    return Response(
        {
            "is_avatar_locked": int(profile.is_avatar_locked),
            "avatar_lock_user_message": profile.avatar_lock_user_message,
            "avatar_lock_staff_message": profile.avatar_lock_staff_message,
        }
    )



from .moderation import *
from .threadparticipant import *
from .thread import *
from .post import *
from .feed import *
from .postedit import *
from .postlike import *
from .attachment import *
from .poll import *
from .pollvote import *



from django.utils.translation import gettext as _

from . import PostingEndpoint, PostingMiddleware
from ....acl import useracl
from ....core.mail import build_mail, send_messages
from ...permissions import can_see_post, can_see_thread


class EmailNotificationMiddleware(PostingMiddleware):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.previous_last_post_on = self.thread.last_post_on

    def use_this_middleware(self):
        return self.mode == PostingEndpoint.REPLY

    def post_save(self, serializer):
        queryset = (
            self.thread.subscription_set.filter(
                send_email=True, last_read_on__gte=self.previous_last_post_on
            )
            .exclude(user=self.user)
            .select_related("user")
        )

        notifications = []
        for subscription in queryset.iterator():
            if self.subscriber_can_see_post(subscription.user):
                notifications.append(self.build_mail(subscription.user))

        if notifications:
            send_messages(notifications)

    def subscriber_can_see_post(self, subscriber):
        user_acl = useracl.get_user_acl(subscriber, self.request.cache_versions)
        see_thread = can_see_thread(user_acl, self.thread)
        see_post = can_see_post(user_acl, self.post)
        return see_thread and see_post

    def build_mail(self, subscriber):
        if subscriber.id == self.thread.starter_id:
            subject = _('%(user)s has replied to your thread "%(thread)s"')
        else:
            subject = _(
                '%(user)s has replied to thread "%(thread)s" that you are watching'
            )

        subject_formats = {"user": self.user.username, "thread": self.thread.title}

        return build_mail(
            subscriber,
            subject % subject_formats,
            "misago/emails/thread/reply",
            sender=self.user,
            context={
                "settings": self.request.settings,
                "thread": self.thread,
                "post": self.post,
            },
        )



from django.urls import reverse


admin_link = reverse("misago:admin:settings:socialauth:index")


def test_providers_list_renders(admin_client):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_providers_list_renders_with_active_provider(admin_client, provider):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_providers_list_renders_with_disabled_provider(admin_client, disabled_provider):
    response = admin_client.get(admin_link)
    assert response.status_code == 200



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import (
    DisableSocialAuthProvider,
    EditSocialAuthProvider,
    MoveDownSocialAuthProvider,
    MoveUpSocialAuthProvider,
    SocialAuthProvidersList,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Social auth providers
        urlpatterns.namespace("social-auth/", "socialauth", "settings")
        urlpatterns.patterns(
            "settings:socialauth",
            path("", SocialAuthProvidersList.as_view(), name="index"),
            path(
                "edit/<slug:pk>/",
                EditSocialAuthProvider.as_view(),
                name="edit",
            ),
            path(
                "down/<slug:pk>/",
                MoveDownSocialAuthProvider.as_view(),
                name="down",
            ),
            path("up/<slug:pk>/", MoveUpSocialAuthProvider.as_view(), name="up"),
            path(
                "disable/<slug:pk>/",
                DisableSocialAuthProvider.as_view(),
                name="disable",
            ),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Social login"),
            description=_(
                "Enable users to sign on and login using their social profile."
            ),
            parent="settings",
            namespace="socialauth",
        )



from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase


def index_post(post):
    if post.id == post.thread.first_post_id:
        post.set_search_document(post.thread.title)
    else:
        post.set_search_document()
    post.save(update_fields=["search_document"])

    post.update_search_vector()
    post.save(update_fields=["search_vector"])


class SearchApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")

        self.api_link = reverse("misago:api:search")

    def test_no_query(self):
        """api handles no search query"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_empty_query(self):
        """api handles empty search query"""
        response = self.client.get("%s?q=" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_short_query(self):
        """api handles short search query"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(thread, message="Lorem ipsum dolor.")
        index_post(post)

        response = self.client.get("%s?q=ip" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_wrong_query(self):
        """api handles query miss"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(thread, message="Lorem ipsum dolor.")
        index_post(post)

        response = self.client.get("%s?q=elit" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_hidden_post(self):
        """hidden posts are extempt from search"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(thread, message="Lorem ipsum dolor.", is_hidden=True)
        index_post(post)

        response = self.client.get("%s?q=ipsum" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_unapproved_post(self):
        """unapproves posts are extempt from search"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(
            thread, message="Lorem ipsum dolor.", is_unapproved=True
        )
        index_post(post)

        response = self.client.get("%s?q=ipsum" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                self.assertEqual(provider["results"]["results"], [])

    def test_query(self):
        """api handles search query"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(thread, message="Lorem ipsum dolor.")
        index_post(post)

        response = self.client.get("%s?q=ipsum" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], post.id)

    def test_thread_title_search(self):
        """api searches threads by title"""
        thread = test.post_thread(self.category, title="Atmosphere of mars")
        index_post(thread.first_post)

        post = test.reply_thread(thread, message="Lorem ipsum dolor.")
        index_post(post)

        response = self.client.get("%s?q=mars atmosphere" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], thread.first_post_id)

    def test_complex_query(self):
        """api handles complex query that uses fulltext search facilities"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(thread, message="Atmosphere of Mars")
        index_post(post)

        response = self.client.get("%s?q=Mars atmosphere" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], post.id)

    def test_filtered_query(self):
        """search filters are used by search system"""
        thread = test.post_thread(self.category)
        post = test.reply_thread(
            thread, message="You just do MMM in 4th minute and its pwnt"
        )

        index_post(post)

        response = self.client.get("%s?q=MMM" % self.api_link)
        self.assertEqual(response.status_code, 200)

        reponse_json = response.json()
        self.assertIn("threads", [p["id"] for p in reponse_json])

        for provider in reponse_json:
            if provider["id"] == "threads":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], post.id)

        response = self.client.get("%s?q=Marines Medics" % self.api_link)
        self.assertEqual(response.status_code, 200)

        for provider in reponse_json:
            if provider["id"] == "threads":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], post.id)


class SearchProviderApiTests(SearchApiTests):
    def setUp(self):
        super().setUp()

        self.api_link = reverse(
            "misago:api:search", kwargs={"search_provider": "threads"}
        )


def test_post_search_filters_hook_is_used_by_threads_search(
    db, user_client, mocker, thread
):
    def search_filter(search):
        return search.replace("apple phone", "iphone")

    mocker.patch(
        "misago.threads.filtersearch.hooks.post_search_filters", [search_filter]
    )

    post = test.reply_thread(thread, message="Lorem ipsum iphone dolor met.")
    index_post(post)

    response = user_client.get("/api/search/?q=apple phone")

    reponse_json = response.json()
    assert "threads" in [p["id"] for p in reponse_json]
    for provider in reponse_json:
        if provider["id"] == "threads":
            results = provider["results"]["results"]
            assert len(results) == 1
            assert results[0]["id"] == post.id



import time
from datetime import timedelta

from django.core.management.base import BaseCommand
from django.utils import timezone

from ....conf.shortcuts import get_dynamic_settings
from ....core.management.progressbar import show_progress
from ....core.pgutils import chunk_queryset
from ...models import Attachment


class Command(BaseCommand):
    help = "Deletes attachments unassociated with any posts"

    def handle(self, *args, **options):
        settings = get_dynamic_settings()

        cutoff = timezone.now() - timedelta(hours=settings.unused_attachments_lifetime)
        queryset = Attachment.objects.filter(post__isnull=True, uploaded_on__lt=cutoff)

        attachments_to_sync = queryset.count()

        if not attachments_to_sync:
            self.stdout.write("\n\nNo unused attachments were cleared")
        else:
            self.sync_attachments(queryset, attachments_to_sync)

    def sync_attachments(self, queryset, attachments_to_sync):
        self.stdout.write("Clearing %s attachments...\n" % attachments_to_sync)

        cleared_count = 0
        show_progress(self, cleared_count, attachments_to_sync)
        start_time = time.time()

        for attachment in chunk_queryset(queryset):
            attachment.delete()

            cleared_count += 1
            show_progress(self, cleared_count, attachments_to_sync, start_time)

        self.stdout.write("\n\nCleared %s attachments" % cleared_count)



from django.db import models
from django.utils.translation import gettext as _

from .cache import clear_acl_cache


def permissions_default():
    return {}


class BaseRole(models.Model):
    name = models.CharField(max_length=255)
    special_role = models.CharField(max_length=255, null=True, blank=True)
    permissions = models.JSONField(default=permissions_default)

    class Meta:
        abstract = True

    def __str__(self):
        return _(self.name)

    def save(self, *args, **kwargs):
        if self.pk:
            clear_acl_cache()
        return super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        clear_acl_cache()
        return super().delete(*args, **kwargs)


class Role(BaseRole):
    pass



import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models

import misago.users.avatars.store


class Migration(migrations.Migration):

    initial = True

    dependencies = [("auth", "0001_initial"), ("misago_acl", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        null=True, blank=True, verbose_name="last login"
                    ),
                ),
                ("username", models.CharField(max_length=30)),
                ("slug", models.CharField(unique=True, max_length=30)),
                ("email", models.EmailField(max_length=255, db_index=True)),
                ("email_hash", models.CharField(unique=True, max_length=32)),
                (
                    "joined_on",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="joined on"
                    ),
                ),
                ("joined_from_ip", models.GenericIPAddressField()),
                ("last_ip", models.GenericIPAddressField(null=True, blank=True)),
                ("is_hiding_presence", models.BooleanField(default=False)),
                ("title", models.CharField(max_length=255, null=True, blank=True)),
                ("requires_activation", models.PositiveIntegerField(default=0)),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into admin sites.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                ("acl_key", models.CharField(max_length=12, null=True, blank=True)),
                (
                    "is_active",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        verbose_name="active",
                        help_text=(
                            "Designates whether this user should be treated as active. Unselect this instead of deleting "
                            "accounts."
                        ),
                    ),
                ),
                ("is_active_staff_message", models.TextField(null=True, blank=True)),
                (
                    "groups",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Group",
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of his/her group.",
                        verbose_name="groups",
                    ),
                ),
                ("roles", models.ManyToManyField(to="misago_acl.Role")),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Permission",
                        blank=True,
                        help_text="Specific permissions for this user.",
                        verbose_name="user permissions",
                    ),
                ),
                (
                    "avatar_tmp",
                    models.ImageField(
                        max_length=255,
                        upload_to=misago.users.avatars.store.upload_to,
                        null=True,
                        blank=True,
                    ),
                ),
                (
                    "avatar_src",
                    models.ImageField(
                        max_length=255,
                        upload_to=misago.users.avatars.store.upload_to,
                        null=True,
                        blank=True,
                    ),
                ),
                (
                    "avatar_crop",
                    models.CharField(max_length=255, null=True, blank=True),
                ),
                ("avatars", models.JSONField(null=True, blank=True)),
                ("is_avatar_locked", models.BooleanField(default=False)),
                ("avatar_lock_user_message", models.TextField(null=True, blank=True)),
                ("avatar_lock_staff_message", models.TextField(null=True, blank=True)),
                ("signature", models.TextField(null=True, blank=True)),
                ("signature_parsed", models.TextField(null=True, blank=True)),
                (
                    "signature_checksum",
                    models.CharField(max_length=64, null=True, blank=True),
                ),
                ("is_signature_locked", models.BooleanField(default=False)),
                (
                    "signature_lock_user_message",
                    models.TextField(null=True, blank=True),
                ),
                (
                    "signature_lock_staff_message",
                    models.TextField(null=True, blank=True),
                ),
                ("following", models.PositiveIntegerField(default=0)),
                ("followers", models.PositiveIntegerField(default=0)),
                (
                    "limits_private_thread_invites_to",
                    models.PositiveIntegerField(default=0),
                ),
                ("unread_private_threads", models.PositiveIntegerField(default=0)),
                ("sync_unread_private_threads", models.BooleanField(default=False)),
                (
                    "subscribe_to_started_threads",
                    models.PositiveIntegerField(default=0),
                ),
                (
                    "subscribe_to_replied_threads",
                    models.PositiveIntegerField(default=0),
                ),
                ("threads", models.PositiveIntegerField(default=0)),
                ("posts", models.PositiveIntegerField(default=0, db_index=True)),
                ("last_posted_on", models.DateTimeField(null=True, blank=True)),
            ],
            options={"abstract": False},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Online",
            fields=[
                ("current_ip", models.GenericIPAddressField()),
                ("last_click", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "user",
                    models.OneToOneField(
                        related_name="online_tracker",
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="UsernameChange",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("changed_by_username", models.CharField(max_length=30)),
                ("changed_on", models.DateTimeField(default=django.utils.timezone.now)),
                ("new_username", models.CharField(max_length=255)),
                ("old_username", models.CharField(max_length=255)),
                (
                    "changed_by",
                    models.ForeignKey(
                        related_name="user_renames",
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        null=True,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        related_name="namechanges",
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"get_latest_by": "changed_on"},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Rank",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("slug", models.CharField(unique=True, max_length=255)),
                ("description", models.TextField(null=True, blank=True)),
                ("title", models.CharField(max_length=255, null=True, blank=True)),
                ("css_class", models.CharField(max_length=255, null=True, blank=True)),
                ("is_default", models.BooleanField(default=False)),
                ("is_tab", models.BooleanField(default=False)),
                ("order", models.IntegerField(default=0)),
                (
                    "roles",
                    models.ManyToManyField(to="misago_acl.Role", null=True, blank=True),
                ),
            ],
            options={"get_latest_by": "order"},
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="user",
            name="rank",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to_field="id",
                blank=True,
                to="misago_users.Rank",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="user",
            name="follows",
            field=models.ManyToManyField(
                related_name="followed_by", to=settings.AUTH_USER_MODEL
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="user",
            name="blocks",
            field=models.ManyToManyField(
                related_name="blocked_by", to=settings.AUTH_USER_MODEL
            ),
            preserve_default=True,
        ),
        migrations.CreateModel(
            name="ActivityRanking",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        related_name="+",
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("score", models.PositiveIntegerField(default=0, db_index=True)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="Avatar",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                ("size", models.PositiveIntegerField(default=0)),
                (
                    "image",
                    models.ImageField(
                        max_length=255, upload_to=misago.users.avatars.store.upload_to
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AvatarGallery",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("gallery", models.CharField(max_length=255)),
                (
                    "image",
                    models.ImageField(
                        max_length=255, upload_to=misago.users.avatars.store.upload_to
                    ),
                ),
            ],
            options={"ordering": ["gallery", "pk"]},
        ),
        migrations.CreateModel(
            name="Ban",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("check_type", models.PositiveIntegerField(default=0, db_index=True)),
                ("banned_value", models.CharField(max_length=255, db_index=True)),
                ("user_message", models.TextField(null=True, blank=True)),
                ("staff_message", models.TextField(null=True, blank=True)),
                (
                    "expires_on",
                    models.DateTimeField(null=True, blank=True, db_index=True),
                ),
                ("is_checked", models.BooleanField(default=True, db_index=True)),
            ],
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="BanCache",
            fields=[
                ("user_message", models.TextField(null=True, blank=True)),
                ("staff_message", models.TextField(null=True, blank=True)),
                ("bans_version", models.PositiveIntegerField(default=0)),
                ("expires_on", models.DateTimeField(null=True, blank=True)),
                (
                    "ban",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.SET_NULL,
                        blank=True,
                        to="misago_users.Ban",
                        null=True,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        related_name="ban_cache",
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



from django.urls import reverse

from .searchproviders import searchproviders


def search_providers(request):
    allowed_providers = []

    try:
        if request.user_acl["can_search"]:
            allowed_providers = searchproviders.get_allowed_providers(request)
    except AttributeError:
        # is user has no acl_cache attribute, cease entire middleware
        # this is edge case that occurs when debug toolbar intercepts
        # the redirect response from logout page and runs context providers
        # with non-misago's anonymous user model that has no acl support
        return {}

    request.frontend_context["SEARCH_URL"] = reverse("misago:search")
    request.frontend_context["SEARCH_API"] = reverse("misago:api:search")
    request.frontend_context["SEARCH_PROVIDERS"] = []

    for provider in allowed_providers:
        request.frontend_context["SEARCH_PROVIDERS"].append(
            {
                "id": provider.url,
                "name": str(provider.name),
                "icon": provider.icon,
                "url": reverse(
                    "misago:search", kwargs={"search_provider": provider.url}
                ),
                "api": reverse(
                    "misago:api:search", kwargs={"search_provider": provider.url}
                ),
                "results": None,
                "time": None,
            }
        )

    return {}



# Generated by Django 3.2.15 on 2023-01-04 12:36

from django.db import migrations

from ..hydrators import dehydrate_value

settings = [
    {
        "setting": "enable_oauth2_client",
        "python_type": "bool",
        "dry_value": False,
        "is_public": True,
    },
    {"setting": "oauth2_provider", "is_public": True},
    {"setting": "oauth2_client_id", "is_public": False},
    {"setting": "oauth2_client_secret", "is_public": False},
    {"setting": "oauth2_scopes", "is_public": False},
    {"setting": "oauth2_login_url", "is_public": False},
    {"setting": "oauth2_token_url", "is_public": False},
    {"setting": "oauth2_token_method", "dry_value": "POST", "is_public": False},
    {"setting": "oauth2_token_extra_headers", "is_public": False},
    {
        "setting": "oauth2_json_token_path",
        "dry_value": "access_token",
        "is_public": False,
    },
    {"setting": "oauth2_user_url", "is_public": False},
    {"setting": "oauth2_user_method", "dry_value": "GET", "is_public": False},
    {
        "setting": "oauth2_user_token_location",
        "dry_value": "QUERY",
        "is_public": False,
    },
    {
        "setting": "oauth2_user_token_name",
        "dry_value": "access_token",
        "is_public": False,
    },
    {"setting": "oauth2_user_extra_headers", "is_public": False},
    {
        "setting": "oauth2_send_welcome_email",
        "python_type": "bool",
        "dry_value": False,
        "is_public": False,
    },
    {"setting": "oauth2_json_id_path", "dry_value": "id", "is_public": False},
    {"setting": "oauth2_json_name_path", "dry_value": "name", "is_public": False},
    {"setting": "oauth2_json_email_path", "dry_value": "email", "is_public": False},
    {"setting": "oauth2_json_avatar_path", "is_public": False},
]


def create_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    for setting in settings:
        data = setting.copy()
        if "python_type" in data and "dry_value" in data:
            data["dry_value"] = dehydrate_value(data["python_type"], data["dry_value"])

        Setting.objects.create(**setting)


class Migration(migrations.Migration):

    dependencies = [
        ("misago_conf", "0006_add_index_message"),
    ]

    operations = [migrations.RunPython(create_settings)]



from django.http import Http404
from django.shortcuts import render
from django.urls import reverse
from django.views import View

from ...core.shortcuts import get_int_or_404
from ..viewmodels import (
    ForumThreads,
    PrivateThreads,
    PrivateThreadsCategory,
    ThreadsCategory,
    ThreadsRootCategory,
)


class ThreadsList(View):
    category = None
    threads = None

    template_name = None

    def get(self, request, list_type=None, **kwargs):
        start = get_int_or_404(request.GET.get("start", 0))

        category = self.get_category(request, **kwargs)
        threads = self.get_threads(request, category, list_type, start)

        frontend_context = self.get_frontend_context(request, category, threads)
        request.frontend_context.update(frontend_context)

        template_context = self.get_template_context(request, category, threads)
        return render(request, self.template_name, template_context)

    def get_category(self, request, **kwargs):
        return self.category(request, **kwargs)  # pylint: disable=not-callable

    def get_threads(self, request, category, list_type, start):
        return self.threads(  # pylint: disable=not-callable
            request, category, list_type, start
        )

    def get_frontend_context(self, request, category, threads):
        context = self.get_default_frontend_context()

        context.update(category.get_frontend_context())
        context.update(threads.get_frontend_context())

        return context

    def get_default_frontend_context(self):
        return {}

    def get_template_context(self, request, category, threads):
        context = self.get_default_template_context()

        context.update(category.get_template_context())
        context.update(threads.get_template_context())

        return context

    def get_default_template_context(self):
        return {}


class ForumThreadsList(ThreadsList):
    category = ThreadsRootCategory
    threads = ForumThreads

    template_name = "misago/threadslist/threads.html"

    def get_default_frontend_context(self):
        return {"MERGE_THREADS_API": reverse("misago:api:thread-merge")}


class CategoryThreadsList(ForumThreadsList):
    category = ThreadsCategory

    template_name = "misago/threadslist/category.html"

    def get_category(self, request, **kwargs):
        category = super().get_category(request, **kwargs)
        if not category.level:
            raise Http404()  # disallow root category access
        return category


class PrivateThreadsList(ThreadsList):
    category = PrivateThreadsCategory
    threads = PrivateThreads

    template_name = "misago/threadslist/private_threads.html"



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ....test import assert_contains, assert_has_error_message
from ... import THEME_CACHE


@pytest.fixture
def create_link(theme):
    return reverse("misago:admin:themes:new-css-link", kwargs={"pk": theme.pk})


@pytest.fixture
def edit_link(theme, css_link):
    return reverse(
        "misago:admin:themes:edit-css-link",
        kwargs={"pk": theme.pk, "css_pk": css_link.pk},
    )


@pytest.fixture
def data():
    return {"name": "CSS link", "url": "https://example.com/cdn.css"}


def test_css_link_creation_form_is_displayed(admin_client, create_link):
    response = admin_client.get(create_link)
    assert response.status_code == 200
    assert_contains(response, "New CSS link")


def test_css_link_can_be_created(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    assert theme.css.exists()


def test_css_link_is_created_with_entered_name(theme, admin_client, create_link, data):
    admin_client.post(create_link, data)
    assert theme.css.last().name == data["name"]


def test_css_link_name_can_be_descriptive(theme, admin_client, create_link, data):
    data["name"] = "font (from font.hosting.com)"
    admin_client.post(create_link, data)
    assert theme.css.last().name == data["name"]


def test_css_link_creation_fails_if_name_is_not_given(
    theme, admin_client, create_link, data
):
    data["name"] = ""
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_link_creation_fails_if_name_is_already_taken_by_other_css_in_theme(
    theme, admin_client, create_link, data, css
):
    data["name"] = css.name
    admin_client.post(create_link, data)
    assert theme.css.count() == 1


def test_css_link_name_usage_check_passess_if_name_is_used_by_other_theme_css(
    other_theme, admin_client, data, css
):
    create_link = reverse(
        "misago:admin:themes:new-css-link", kwargs={"pk": other_theme.pk}
    )
    data["name"] = css.name
    admin_client.post(create_link, data)
    assert other_theme.css.exists()


def test_css_link_creation_fails_if_url_is_not_given(
    theme, admin_client, create_link, data
):
    data["url"] = ""
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_link_creation_fails_if_url_is_not_valid(
    theme, admin_client, create_link, data
):
    data["url"] = "invalid-url"
    admin_client.post(create_link, data)
    assert not theme.css.exists()


def test_css_link_is_created_with_correct_order(
    theme, admin_client, create_link, css, data
):
    admin_client.post(create_link, data)
    css_link = theme.css.get(name=data["name"])
    assert css_link.order == 1


def test_css_link_creation_queues_task_to_download_remote_css_size(
    theme, admin_client, create_link, data, mock_update_remote_css_size
):
    admin_client.post(create_link, data)
    css_link = theme.css.last()
    mock_update_remote_css_size.assert_called_once_with(css_link.pk)


def test_css_link_creation_invalidates_theme_cache(
    theme, admin_client, create_link, data
):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(create_link, data)


def test_error_message_is_set_if_user_attempts_to_create_css_link_in_default_theme(
    default_theme, admin_client
):
    create_link = reverse(
        "misago:admin:themes:new-css-link", kwargs={"pk": default_theme.pk}
    )
    response = admin_client.get(create_link)
    assert_has_error_message(response)


def test_error_message_is_set_if_user_attempts_to_create_css_link_in_nonexisting_theme(
    nonexisting_theme, admin_client
):
    create_link = reverse(
        "misago:admin:themes:new-css-link", kwargs={"pk": nonexisting_theme.pk}
    )
    response = admin_client.get(create_link)
    assert_has_error_message(response)


def test_css_link_creation_form_redirects_user_to_new_creation_form_after_creation(
    theme, admin_client, create_link, data
):
    data["stay"] = "1"
    response = admin_client.post(create_link, data)
    assert response["location"] == reverse(
        "misago:admin:themes:new-css-link", kwargs={"pk": theme.pk}
    )


def test_css_link_edition_form_is_displayed(admin_client, edit_link, css_link):
    response = admin_client.get(edit_link)
    assert response.status_code == 200
    assert_contains(response, css_link.name)


def test_css_link_name_can_be_changed(admin_client, edit_link, css_link, data):
    data["name"] = "new link name"
    admin_client.post(edit_link, data)

    css_link.refresh_from_db()
    assert css_link.name == data["name"]


def test_css_link_url_can_be_changed(admin_client, edit_link, css_link, data):
    data["url"] = "https://new.css-link.com/test.css"
    admin_client.post(edit_link, data)

    css_link.refresh_from_db()
    assert css_link.url == data["url"]


def test_changing_css_link_url_queues_task_to_download_remote_css_size(
    admin_client, edit_link, css_link, data, mock_update_remote_css_size
):
    data["url"] = "https://new.css-link.com/test.css"
    admin_client.post(edit_link, data)
    css_link.refresh_from_db()
    mock_update_remote_css_size.assert_called_once_with(css_link.pk)


def test_not_changing_css_link_url_doesnt_queue_task_to_download_remote_css_size(
    admin_client, edit_link, css_link, data, mock_update_remote_css_size
):
    admin_client.post(edit_link, data)
    css_link.refresh_from_db()
    mock_update_remote_css_size.assert_not_called()


def test_changing_css_link_invalidates_theme_cache(admin_client, edit_link, data):
    with assert_invalidates_cache(THEME_CACHE):
        admin_client.post(edit_link, data)


def test_css_order_stays_the_same_after_edit(admin_client, edit_link, css_link, data):
    original_order = css_link.order
    data["name"] = "changed link"
    admin_client.post(edit_link, data)

    css_link.refresh_from_db()
    assert css_link.order == original_order


def test_error_message_is_set_if_user_attempts_to_edit_css_file_with_link_form(
    theme, admin_client, css
):
    edit_link = reverse(
        "misago:admin:themes:edit-css-link", kwargs={"pk": theme.pk, "css_pk": css.pk}
    )
    response = admin_client.get(edit_link)
    assert_has_error_message(response)



from django.http import HttpResponse

from ..middleware import RealIPMiddleware


def get_response(*_):
    return HttpResponse("OK!")


class MockRequest:
    def __init__(self, addr, forwarded_for=None):
        self.META = {"REMOTE_ADDR": addr}

        if forwarded_for:
            self.META["HTTP_X_FORWARDED_FOR"] = forwarded_for


def test_user_ip_middleware_sets_ip_from_remote_add_on_request():
    request = MockRequest("83.42.13.77")
    RealIPMiddleware(get_response)(request)

    assert request.user_ip == request.META["REMOTE_ADDR"]


def test_user_ip_middleware_sets_ip_from_forwarded_for_on_request():
    request = MockRequest("127.0.0.1", "83.42.13.77")
    RealIPMiddleware(get_response)(request)

    assert request.user_ip == request.META["HTTP_X_FORWARDED_FOR"]



from math import ceil

from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext as _
from django.views import View

from ...conf import settings
from ...readtracker.cutoffdate import get_cutoff_date
from ..permissions import exclude_invisible_posts
from ..viewmodels import ForumThread, PrivateThread


class GotoView(View):
    thread = None

    def get(self, request, pk, slug, **kwargs):
        thread = self.get_thread(request, pk, slug).unwrap()
        self.test_permissions(request, thread)

        posts_queryset = exclude_invisible_posts(
            request.user_acl, thread.category, thread.post_set
        )

        target_post = self.get_target_post(
            request.user, thread, posts_queryset.order_by("id"), **kwargs
        )
        target_page = self.compute_post_page(target_post, posts_queryset)

        return self.get_redirect(thread, target_post, target_page)

    def get_thread(self, request, pk, slug):
        return self.thread(request, pk, slug)  # pylint: disable=not-callable

    def test_permissions(self, request, thread):
        pass

    def get_target_post(self, user, thread, posts_queryset):
        raise NotImplementedError(
            "goto views should define their own get_target_post method"
        )

    def compute_post_page(self, target_post, posts_queryset):
        # filter out events, order queryset
        posts_queryset = posts_queryset.filter(is_event=False).order_by("id")
        thread_length = posts_queryset.count()

        # is target an event?
        if target_post.is_event:
            target_event = target_post
            previous_posts = posts_queryset.filter(id__lt=target_event.id)
        else:
            previous_posts = posts_queryset.filter(id__lte=target_post.id)

        post_position = previous_posts.count()

        per_page = self.request.settings.posts_per_page - 1
        orphans = self.request.settings.posts_per_page_orphans
        if orphans:
            orphans += 1

        hits = max(1, thread_length - orphans)
        thread_pages = int(ceil(hits / float(per_page)))

        if post_position >= thread_pages * per_page:
            return thread_pages

        return int(ceil(float(post_position) / (per_page)))

    def get_redirect(self, thread, target_post, target_page):
        thread_url = thread.thread_type.get_thread_absolute_url(thread, target_page)
        return redirect("%s#post-%s" % (thread_url, target_post.pk))


class ThreadGotoPostView(GotoView):
    thread = ForumThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return get_object_or_404(posts_queryset, pk=kwargs["post"])


class ThreadGotoLastView(GotoView):
    thread = ForumThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return posts_queryset.order_by("id").last()


class GetFirstUnreadPostMixin:
    def get_first_unread_post(self, user, posts_queryset):
        if user.is_authenticated:
            cutoff_date = get_cutoff_date(self.request.settings, user)
            expired_posts = Q(posted_on__lt=cutoff_date)
            read_posts = Q(id__in=user.postread_set.values("post"))

            first_unread = (
                posts_queryset.exclude(expired_posts | read_posts)
                .order_by("id")
                .first()
            )

            if first_unread:
                return first_unread

        return posts_queryset.order_by("id").last()


class ThreadGotoNewView(GotoView, GetFirstUnreadPostMixin):
    thread = ForumThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return self.get_first_unread_post(user, posts_queryset)


class ThreadGotoBestAnswerView(GotoView):
    thread = ForumThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return thread.best_answer or thread.first_post


class ThreadGotoUnapprovedView(GotoView):
    thread = ForumThread

    def test_permissions(self, request, thread):
        if not thread.acl["can_approve"]:
            raise PermissionDenied(
                _(
                    "You need permission to approve content to "
                    "be able to go to first unapproved post."
                )
            )

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        unapproved_post = (
            posts_queryset.filter(is_unapproved=True).order_by("id").first()
        )
        if unapproved_post:
            return unapproved_post
        return posts_queryset.order_by("id").last()


class PrivateThreadGotoPostView(GotoView):
    thread = PrivateThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return get_object_or_404(posts_queryset, pk=kwargs["post"])


class PrivateThreadGotoLastView(GotoView):
    thread = PrivateThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return posts_queryset.order_by("id").last()


class PrivateThreadGotoNewView(GotoView, GetFirstUnreadPostMixin):
    thread = PrivateThread

    def get_target_post(self, user, thread, posts_queryset, **kwargs):
        return self.get_first_unread_post(user, posts_queryset)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_multi_line_spoiler 1"
] = """<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Sit amet elit.</p>
<p>Another line.</p>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>"""

snapshots[
    "test_single_line_spoiler 1"
] = """<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Daenerys and Jon live happily ever after!</p>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>"""

snapshots[
    "test_spoiler_can_contain_bbcode_or_markdown 1"
] = """<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Sit <strong>amet</strong> <u>elit</u>.</p>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>"""

snapshots[
    "test_spoilers_can_be_nested 1"
] = """<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Sit amet elit.</p>
<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Nested spoiler</p>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>"""

snapshots[
    "test_spoilers_can_contain_hr_markdown 1"
] = """<aside class="spoiler-block">
<blockquote class="spoiler-body" data-block="spoiler">
<p>Sit amet elit.</p>
<hr />
<p>Another line.</p>
</blockquote>
<div class="spoiler-overlay" data-noquote="1"><button class="spoiler-reveal" type="button"></button></div>
</aside>"""



from unittest.mock import Mock

from ..context_processors import preload_socialauth_json


def test_context_processor_sets_socialauth_entry_in_frontend_context():
    request = Mock(frontend_context={}, socialauth={})
    preload_socialauth_json(request)
    assert "SOCIAL_AUTH" in request.frontend_context



import pytest
from django.urls import reverse

from ....test import assert_contains, assert_not_contains, assert_has_error_message


@pytest.fixture
def assets_client(admin_client):
    def get_theme_assets(theme):
        url = reverse("misago:admin:themes:assets", kwargs={"pk": theme.pk})
        return admin_client.get(url)

    return get_theme_assets


def test_theme_assets_list_is_displayed(assets_client, theme):
    response = assets_client(theme)
    assert_contains(response, theme.name)


def test_css_file_is_displayed_on_theme_asset_list(assets_client, theme, css):
    response = assets_client(theme)
    assert_contains(response, css.name)


def test_css_link_is_displayed_on_theme_asset_list(assets_client, theme, css_link):
    response = assets_client(theme)
    assert_contains(response, css_link.name)


def test_media_is_displayed_on_themes_asset_list(assets_client, theme, media):
    response = assets_client(theme)
    assert_contains(response, media.name)


def test_image_is_displayed_on_themes_asset_list(assets_client, theme, image):
    response = assets_client(theme)
    assert_contains(response, image.name)


def test_image_thumbnail_is_displayed_on_themes_asset_list(assets_client, theme, image):
    response = assets_client(theme)
    assert_contains(response, image.thumbnail.url)


def test_other_theme_assets_are_not_displayed(
    assets_client, other_theme, css, css_link, media, image
):
    response = assets_client(other_theme)
    assert_not_contains(response, css.name)
    assert_not_contains(response, css_link.name)
    assert_not_contains(response, media.name)
    assert_not_contains(response, image.name)


def test_user_is_redirected_away_with_message_from_default_theme_assets_list(
    assets_client, default_theme
):
    response = assets_client(default_theme)
    assert response.status_code == 302
    assert_has_error_message(response)


def test_user_is_redirected_away_with_message_from_nonexisting_theme(
    assets_client, nonexisting_theme
):
    response = assets_client(nonexisting_theme)
    assert response.status_code == 302
    assert_has_error_message(response)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_strikethrough_markdown 1"] = "<p>Lorem <del>ipsum</del> dolor met!</p>"



from django.test import TestCase
from rest_framework.exceptions import ValidationError

from .. import test
from ...categories.models import Category
from ...users.test import create_test_user
from ..mergeconflict import MergeConflict


class MergeConflictTests(TestCase):
    def setUp(self):
        self.category = Category.objects.get(slug="first-category")
        self.user = create_test_user("User", "user@example.com")

    def create_plain_thread(self):
        return test.post_thread(self.category)

    def create_poll_thread(self):
        thread = test.post_thread(self.category)
        test.post_poll(thread, self.user)
        return thread

    def create_best_answer_thread(self):
        thread = test.post_thread(self.category)
        best_answer = test.reply_thread(thread)
        thread.set_best_answer(self.user, best_answer)
        thread.synchronize()
        thread.save()
        return thread

    def test_plain_threads_no_conflicts(self):
        """threads without items of interest don't conflict"""
        threads = [self.create_plain_thread() for i in range(10)]
        merge_conflict = MergeConflict(threads=threads)
        self.assertFalse(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), [])

        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(merge_conflict.get_resolution(), {})

    def test_one_best_answer_one_plain(self):
        """thread with best answer and plain thread don't conflict"""
        threads = [self.create_best_answer_thread(), self.create_plain_thread()]
        merge_conflict = MergeConflict(threads=threads)
        self.assertFalse(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), [])

        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(merge_conflict.get_resolution(), {"best_answer": threads[0]})

    def test_one_poll_one_plain(self):
        """thread with poll and plain thread don't conflict"""
        threads = [self.create_poll_thread(), self.create_plain_thread()]
        merge_conflict = MergeConflict(threads=threads)
        self.assertFalse(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), [])

        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(merge_conflict.get_resolution(), {"poll": threads[0].poll})

    def test_one_best_answer_one_poll(self):
        """thread with best answer and thread with poll don't conflict"""
        threads = [self.create_poll_thread(), self.create_best_answer_thread()]
        merge_conflict = MergeConflict(threads=threads)
        self.assertFalse(merge_conflict.is_merge_conflict())

    def test_one_best_answer_one_poll_one_plain(self):
        """thread with best answer, thread with poll and plain thread don't conflict"""
        threads = [
            self.create_plain_thread(),
            self.create_poll_thread(),
            self.create_best_answer_thread(),
        ]
        merge_conflict = MergeConflict(threads=threads)
        self.assertFalse(merge_conflict.is_merge_conflict())

    def test_three_best_answers_one_poll_two_plain_conflict(self):
        """
        three threads with best answer, thread with poll and two plain threads conflict
        """
        best_answers = [self.create_best_answer_thread() for i in range(3)]
        polls = [self.create_poll_thread()]
        threads = (
            [self.create_plain_thread(), self.create_plain_thread()]
            + best_answers
            + polls
        )

        merge_conflict = MergeConflict(threads=threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["best_answer"])

        # without choice, conflict lists resolutions
        try:
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(merge_conflict.get_conflicting_fields(), ["best_answer"])
            self.assertEqual(
                e.detail,
                {
                    "best_answers": [["0", "Unmark all best answers"]]
                    + [[str(thread.id), thread.title] for thread in best_answers]
                },
            )

        # conflict validates choice
        try:
            merge_conflict = MergeConflict({"best_answer": threads[0].id}, threads)
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(e.detail, {"best_answer": ["Invalid choice."]})

        # conflict returns selected resolution
        merge_conflict = MergeConflict({"best_answer": best_answers[0].id}, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["best_answer"])
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": best_answers[0], "poll": polls[0].poll},
        )

        # conflict returns no-choice resolution
        merge_conflict = MergeConflict({"best_answer": 0}, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["best_answer"])
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": None, "poll": polls[0].poll},
        )

    def test_one_best_answer_three_polls_two_plain_conflict(self):
        """
        one thread with best answer, three threads with poll
        and two plain threads conflict
        """
        best_answers = [self.create_best_answer_thread()]
        polls = [self.create_poll_thread() for i in range(3)]
        threads = (
            [self.create_plain_thread(), self.create_plain_thread()]
            + best_answers
            + polls
        )

        merge_conflict = MergeConflict(threads=threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["poll"])

        # without choice, conflict lists resolutions
        try:
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(merge_conflict.get_conflicting_fields(), ["poll"])
            self.assertEqual(
                e.detail,
                {
                    "polls": [["0", "Delete all polls"]]
                    + [
                        [
                            str(thread.poll.id),
                            "%s (%s)" % (thread.poll.question, thread.title),
                        ]
                        for thread in polls
                    ]
                },
            )

        # conflict validates choice
        try:
            merge_conflict = MergeConflict({"poll": threads[0].id}, threads)
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(e.detail, {"poll": ["Invalid choice."]})

        # conflict returns selected resolution
        merge_conflict = MergeConflict({"poll": polls[0].poll.id}, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["poll"])
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": best_answers[0], "poll": polls[0].poll},
        )

        # conflict returns no-choice resolution
        merge_conflict = MergeConflict({"poll": 0}, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(merge_conflict.get_conflicting_fields(), ["poll"])
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": best_answers[0], "poll": None},
        )

    def test_three_best_answers_three_polls_two_plain_conflict(self):
        """multiple conflict is handled"""
        best_answers = [self.create_best_answer_thread() for i in range(3)]
        polls = [self.create_poll_thread() for i in range(3)]
        threads = (
            [self.create_plain_thread(), self.create_plain_thread()]
            + best_answers
            + polls
        )

        merge_conflict = MergeConflict(threads=threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(
            merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
        )

        # without choice, conflict lists all resolutions
        try:
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(
                merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
            )
            self.assertEqual(
                e.detail,
                {
                    "best_answers": [["0", "Unmark all best answers"]]
                    + [[str(thread.id), thread.title] for thread in best_answers],
                    "polls": [["0", "Delete all polls"]]
                    + [
                        [
                            str(thread.poll.id),
                            "%s (%s)" % (thread.poll.question, thread.title),
                        ]
                        for thread in polls
                    ],
                },
            )

        # conflict validates all choices if single choice was given
        try:
            merge_conflict = MergeConflict({"best_answer": threads[0].id}, threads)
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(
                e.detail,
                {"best_answer": ["Invalid choice."], "poll": ["Invalid choice."]},
            )

        try:
            merge_conflict = MergeConflict({"poll": threads[0].id}, threads)
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(
                e.detail,
                {"best_answer": ["Invalid choice."], "poll": ["Invalid choice."]},
            )

        # conflict validates all choices if all choices were given
        try:
            merge_conflict = MergeConflict(
                {"best_answer": threads[0].id, "poll": threads[0].id}, threads
            )
            merge_conflict.is_valid(raise_exception=True)
            self.fail("merge_conflict.is_valid() should raise ValidationError")
        except ValidationError as e:
            self.assertTrue(merge_conflict.is_merge_conflict())
            self.assertEqual(
                e.detail,
                {"best_answer": ["Invalid choice."], "poll": ["Invalid choice."]},
            )

        # conflict returns selected resolutions
        valid_choices = {"best_answer": best_answers[0].id, "poll": polls[0].poll.id}
        merge_conflict = MergeConflict(valid_choices, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(
            merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
        )
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": best_answers[0], "poll": polls[0].poll},
        )

        # conflict returns no-choice resolution
        merge_conflict = MergeConflict({"best_answer": 0, "poll": 0}, threads)
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(
            merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
        )
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(), {"best_answer": None, "poll": None}
        )

        # conflict allows mixing no-choice with choice
        merge_conflict = MergeConflict(
            {"best_answer": best_answers[0].id, "poll": 0}, threads
        )
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(
            merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
        )
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": best_answers[0], "poll": None},
        )

        merge_conflict = MergeConflict(
            {"best_answer": 0, "poll": polls[0].poll.id}, threads
        )
        self.assertTrue(merge_conflict.is_merge_conflict())
        self.assertEqual(
            merge_conflict.get_conflicting_fields(), ["best_answer", "poll"]
        )
        merge_conflict.is_valid(raise_exception=True)
        self.assertEqual(
            merge_conflict.get_resolution(),
            {"best_answer": None, "poll": polls[0].poll},
        )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_code_block_disables_parsing 1"
] = "<pre><code>Dolor [b]met.[/b]</code></pre>"

snapshots[
    "test_code_with_language_parameter 1"
] = '<pre><code class="language-php">echo(&quot;Hello!&quot;);</code></pre>'

snapshots[
    "test_code_with_quoted_language_parameter 1"
] = '<pre><code class="language-php">echo(&quot;Hello!&quot;);</code></pre>'

snapshots[
    "test_multi_line_code 1"
] = """<pre><code>&lt;script&gt;
alert(&quot;!&quot;)
&lt;/script&gt;</code></pre>"""

snapshots[
    "test_single_line_code 1"
] = "<pre><code>echo(&quot;Hello!&quot;);</code></pre>"



from django.core.exceptions import PermissionDenied
from rest_framework.response import Response

from ....acl.objectacl import add_acl_to_obj
from ...events import record_event
from ...mergeconflict import MergeConflict
from ...models import Thread
from ...moderation import threads as moderation
from ...permissions import allow_merge_thread
from ...serializers import (
    MergeThreadSerializer,
    MergeThreadsSerializer,
    ThreadsListSerializer,
)


def thread_merge_endpoint(
    request, thread, viewmodel
):  # pylint: disable=too-many-branches
    allow_merge_thread(request.user_acl, thread)

    serializer = MergeThreadSerializer(
        data=request.data,
        context={"request": request, "thread": thread, "viewmodel": viewmodel},
    )

    if not serializer.is_valid():
        if "other_thread" in serializer.errors:
            errors = serializer.errors["other_thread"]
        elif "best_answer" in serializer.errors:
            errors = serializer.errors["best_answer"]
        elif "best_answers" in serializer.errors:
            return Response(
                {"best_answers": serializer.errors["best_answers"]}, status=400
            )
        elif "poll" in serializer.errors:
            errors = serializer.errors["poll"]
        elif "polls" in serializer.errors:
            return Response({"polls": serializer.errors["polls"]}, status=400)
        else:
            errors = list(serializer.errors.values())[0]
        return Response({"detail": errors[0]}, status=400)

    # merge conflict
    other_thread = serializer.validated_data["other_thread"]

    best_answer = serializer.validated_data.get("best_answer")
    if "best_answer" in serializer.merge_conflict and not best_answer:
        other_thread.clear_best_answer()
    if best_answer and best_answer != other_thread:
        other_thread.best_answer_id = thread.best_answer_id
        other_thread.best_answer_is_protected = thread.best_answer_is_protected
        other_thread.best_answer_marked_on = thread.best_answer_marked_on
        other_thread.best_answer_marked_by_id = thread.best_answer_marked_by_id
        other_thread.best_answer_marked_by_name = thread.best_answer_marked_by_name
        other_thread.best_answer_marked_by_slug = thread.best_answer_marked_by_slug

    poll = serializer.validated_data.get("poll")
    if "poll" in serializer.merge_conflict:
        if poll and poll.thread_id != other_thread.id:
            other_thread.poll.delete()
            poll.move(other_thread)
        elif not poll:
            other_thread.poll.delete()
    elif poll:
        poll.move(other_thread)

    # merge thread contents
    moderation.merge_thread(request, other_thread, thread)

    other_thread.synchronize()
    other_thread.save()

    other_thread.category.synchronize()
    other_thread.category.save()

    if thread.category != other_thread.category:
        thread.category.synchronize()
        thread.category.save()

    return Response(
        {
            "id": other_thread.pk,
            "title": other_thread.title,
            "url": other_thread.get_absolute_url(),
        }
    )


def threads_merge_endpoint(request):
    serializer = MergeThreadsSerializer(
        data=request.data,
        context={"settings": request.settings, "user_acl": request.user_acl},
    )

    if not serializer.is_valid():
        if "threads" in serializer.errors:
            errors = {"detail": serializer.errors["threads"][0]}
            return Response(errors, status=403)
        if "non_field_errors" in serializer.errors:
            errors = {"detail": serializer.errors["non_field_errors"][0]}
            return Response(errors, status=403)
        return Response(serializer.errors, status=400)

    threads = serializer.validated_data["threads"]
    invalid_threads = []

    for thread in threads:
        try:
            allow_merge_thread(request.user_acl, thread)
        except PermissionDenied as e:
            invalid_threads.append(
                {"id": thread.pk, "title": thread.title, "errors": [str(e)]}
            )

    if invalid_threads:
        return Response(invalid_threads, status=403)

    # handle merge conflict
    merge_conflict = MergeConflict(serializer.validated_data, threads)
    merge_conflict.is_valid(raise_exception=True)

    new_thread = merge_threads(
        request, serializer.validated_data, threads, merge_conflict
    )
    return Response(ThreadsListSerializer(new_thread).data)


def merge_threads(request, validated_data, threads, merge_conflict):
    new_thread = Thread(
        category=validated_data["category"],
        started_on=threads[0].started_on,
        last_post_on=threads[0].last_post_on,
    )

    new_thread.set_title(validated_data["title"])
    new_thread.save()

    resolution = merge_conflict.get_resolution()

    best_answer = resolution.get("best_answer")
    if best_answer:
        new_thread.best_answer_id = best_answer.best_answer_id
        new_thread.best_answer_is_protected = best_answer.best_answer_is_protected
        new_thread.best_answer_marked_on = best_answer.best_answer_marked_on
        new_thread.best_answer_marked_by_id = best_answer.best_answer_marked_by_id
        new_thread.best_answer_marked_by_name = best_answer.best_answer_marked_by_name
        new_thread.best_answer_marked_by_slug = best_answer.best_answer_marked_by_slug

    poll = resolution.get("poll")
    if poll:
        poll.move(new_thread)

    categories = []
    for thread in threads:
        categories.append(thread.category)
        new_thread.merge(thread)
        thread.delete()

        record_event(
            request, new_thread, "merged", {"merged_thread": thread.title}, commit=False
        )

    new_thread.synchronize()
    new_thread.save()

    if validated_data.get("weight") == Thread.WEIGHT_GLOBAL:
        moderation.pin_thread_globally(request, new_thread)
    elif validated_data.get("weight"):
        moderation.pin_thread_locally(request, new_thread)
    if validated_data.get("is_hidden", False):
        moderation.hide_thread(request, new_thread)
    if validated_data.get("is_closed", False):
        moderation.close_thread(request, new_thread)

    if new_thread.category not in categories:
        categories.append(new_thread.category)

    for category in categories:
        category.synchronize()
        category.save()

    # set extra attrs on thread for UI
    new_thread.is_read = False
    new_thread.subscription = None

    add_acl_to_obj(request.user_acl, new_thread)
    return new_thread



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...threads import test
from ..models import Ban
from ..test import AuthenticatedUserTestCase, create_test_user


class UserProfileViewsTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.link_kwargs = {"slug": self.user.slug, "pk": self.user.pk}

        self.category = Category.objects.get(slug="first-category")

    def test_outdated_slugs(self):
        """user profile view redirects to valid slug"""
        response = self.client.get(
            reverse("misago:user-posts", kwargs={"slug": "baww", "pk": self.user.pk})
        )

        self.assertEqual(response.status_code, 301)

    def test_user_disabled(self):
        """disabled user's profile returns 404 for non-admins"""
        self.user.is_staff = False
        self.user.save()

        disabled_user = create_test_user("DisabledUser", "disabled@example.com")

        disabled_user.is_active = False
        disabled_user.save()

        response = self.client.get(disabled_user.get_absolute_url())
        self.assertEqual(response.status_code, 404)

        self.user.is_staff = True
        self.user.save()

        response = self.client.get(disabled_user.get_absolute_url())
        self.assertEqual(response.status_code, 302)

        # profile page displays notice about user being disabled
        response = self.client.get(response["location"])
        self.assertContains(response, "account has been disabled")

    def test_user_posts_list(self):
        """user profile posts list has no showstoppers"""
        link = reverse("misago:user-posts", kwargs=self.link_kwargs)
        response = self.client.get(link)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "You have posted no messages")

        thread = test.post_thread(category=self.category, poster=self.user)

        response = self.client.get(link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, thread.get_absolute_url())

        post = test.reply_thread(thread, poster=self.user)
        other_post = test.reply_thread(thread, poster=self.user)

        response = self.client.get(link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, post.get_absolute_url())
        self.assertContains(response, other_post.get_absolute_url())

    def test_user_threads_list(self):
        """user profile threads list has no showstoppers"""
        link = reverse("misago:user-threads", kwargs=self.link_kwargs)

        response = self.client.get(link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "You have no started threads.")

        thread = test.post_thread(category=self.category, poster=self.user)

        response = self.client.get(link)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, thread.get_absolute_url())

        post = test.reply_thread(thread, poster=self.user)
        other_post = test.reply_thread(thread, poster=self.user)

        response = self.client.get(link)
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, post.get_absolute_url())
        self.assertNotContains(response, other_post.get_absolute_url())

    def test_user_followers(self):
        """user profile followers list has no showstoppers"""
        response = self.client.get(
            reverse("misago:user-followers", kwargs=self.link_kwargs)
        )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "You have no followers.")

        followers = []
        for i in range(10):
            user_data = ("Follower%s" % i, "foll%s@test.com" % i)
            followers.append(create_test_user(*user_data))
            self.user.followed_by.add(followers[-1])

        response = self.client.get(
            reverse("misago:user-followers", kwargs=self.link_kwargs)
        )
        self.assertEqual(response.status_code, 200)
        for i in range(10):
            self.assertContains(response, "Follower%s" % i)

    def test_user_follows(self):
        """user profile follows list has no showstoppers"""
        response = self.client.get(
            reverse("misago:user-follows", kwargs=self.link_kwargs)
        )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "You are not following any users.")

        followers = []
        for i in range(10):
            user_data = ("Follower%s" % i, "foll%s@test.com" % i)
            followers.append(create_test_user(*user_data))
            followers[-1].followed_by.add(self.user)

        response = self.client.get(
            reverse("misago:user-follows", kwargs=self.link_kwargs)
        )
        self.assertEqual(response.status_code, 200)
        for i in range(10):
            self.assertContains(response, "Follower%s" % i)

    def test_user_details(self):
        """user details page has no showstoppers"""
        response = self.client.get(
            reverse("misago:user-details", kwargs=self.link_kwargs)
        )

        self.assertEqual(response.status_code, 200)

    def test_username_history_list(self):
        """user name changes history list has no showstoppers"""
        response = self.client.get(
            reverse("misago:username-history", kwargs=self.link_kwargs)
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Your username was never changed.")

        self.user.set_username("RenamedAdmin")
        self.user.save()
        self.user.set_username("TestUser")
        self.user.save()

        response = self.client.get(
            reverse("misago:username-history", kwargs=self.link_kwargs)
        )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "TestUser")
        self.assertContains(response, "RenamedAdmin")

    def test_user_ban_details(self):
        """user ban details page has no showstoppers"""
        test_user = create_test_user("User", "user@example.com")
        link_kwargs = {"slug": test_user.slug, "pk": test_user.pk}

        with patch_user_acl({"can_see_ban_details": 0}):
            response = self.client.get(reverse("misago:user-ban", kwargs=link_kwargs))
            self.assertEqual(response.status_code, 404)

        with patch_user_acl({"can_see_ban_details": 1}):
            response = self.client.get(reverse("misago:user-ban", kwargs=link_kwargs))
            self.assertEqual(response.status_code, 404)

        Ban.objects.create(
            banned_value=test_user.username,
            user_message="User m3ss4ge.",
            staff_message="Staff m3ss4ge.",
            is_checked=True,
        )
        test_user.ban_cache.delete()

        with patch_user_acl({"can_see_ban_details": 1}):
            response = self.client.get(reverse("misago:user-ban", kwargs=link_kwargs))

            self.assertEqual(response.status_code, 200)
            self.assertContains(response, "User m3ss4ge")
            self.assertContains(response, "Staff m3ss4ge")



# Generated by Django 2.2.1 on 2019-05-18 16:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_threads", "0011_remove_custom_partial_indexes")]

    operations = [
        migrations.AddIndex(
            model_name="post",
            index=models.Index(
                condition=models.Q(has_open_reports=True),
                fields=["has_open_reports"],
                name="misago_post_has_open_repo_part",
            ),
        ),
        migrations.AddIndex(
            model_name="post",
            index=models.Index(
                condition=models.Q(is_hidden=False),
                fields=["is_hidden"],
                name="misago_post_is_hidden_part",
            ),
        ),
        migrations.AddIndex(
            model_name="post",
            index=models.Index(
                condition=models.Q(is_event=True),
                fields=["is_event", "event_type"],
                name="misago_post_is_event_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(weight=2),
                fields=["weight"],
                name="misago_thread_pinned_glob_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(weight=1),
                fields=["weight"],
                name="misago_thread_pinned_loca_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(weight=0),
                fields=["weight"],
                name="misago_thread_not_pinned_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(weight__lt=2),
                fields=["weight"],
                name="misago_thread_not_global_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(has_reported_posts=True),
                fields=["has_reported_posts"],
                name="misago_thread_has_reporte_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(has_unapproved_posts=True),
                fields=["has_unapproved_posts"],
                name="misago_thread_has_unappro_part",
            ),
        ),
        migrations.AddIndex(
            model_name="thread",
            index=models.Index(
                condition=models.Q(is_hidden=False),
                fields=["is_hidden"],
                name="misago_thread_is_visible_part",
            ),
        ),
    ]



from .providers import providers


def build_acl(roles):
    """build ACL for given roles"""
    acl = {}

    for extension, module in providers.list():
        try:
            acl = module.build_acl(acl, roles, extension)
        except AttributeError:
            message = "%s has to define build_acl function" % extension
            raise AttributeError(message)

    return acl



from django import template

register = template.Library()


@register.filter(name="avatar")
def avatar(user, size=200):
    found_avatar = user.avatars[0]
    for user_avatar in user.avatars:
        if user_avatar["size"] >= size:
            found_avatar = user_avatar
    return found_avatar["url"]



from django.urls import reverse

from ...acl.test import patch_user_acl
from ..test import AuthenticatedUserTestCase, create_test_user


class SearchApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:search")

    @patch_user_acl({"can_search_users": 0})
    def test_no_permission(self):
        """api respects permission to search users"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)
        self.assertNotIn("users", [p["id"] for p in response.json()])

    def test_no_query(self):
        """api handles no search query"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                self.assertEqual(provider["results"]["results"], [])

    def test_empty_query(self):
        """api handles empty search query"""
        response = self.client.get("%s?q=" % self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                self.assertEqual(provider["results"]["results"], [])

    def test_short_query(self):
        """api handles short search query"""
        response = self.client.get("%s?q=%s" % (self.api_link, self.user.username[0]))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], self.user.id)

    def test_exact_match(self):
        """api handles exact search query"""
        response = self.client.get("%s?q=%s" % (self.api_link, self.user.username))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], self.user.id)

    def test_orphans_match(self):
        """api handles last three chars match query"""
        response = self.client.get("%s?q=%s" % (self.api_link, self.user.username[-3:]))
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], self.user.id)

    def test_no_match(self):
        """api handles no match"""
        response = self.client.get("%s?q=BobBoberson" % self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                self.assertEqual(provider["results"]["results"], [])

    def test_search_disabled_user(self):
        """api respects disabled users visibility"""
        disabled_user = create_test_user(
            "DisabledUser", "disableduser@example.com", is_active=False
        )

        response = self.client.get("%s?q=DisabledUser" % self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                self.assertEqual(provider["results"]["results"], [])

        # user shows in searchech performed by staff
        self.user.is_staff = True
        self.user.save()

        response = self.client.get("%s?q=DisabledUser" % self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertIn("users", [p["id"] for p in response_json])

        for provider in response_json:
            if provider["id"] == "users":
                results = provider["results"]["results"]
                self.assertEqual(len(results), 1)
                self.assertEqual(results[0]["id"], disabled_user.id)


class SearchProviderApiTests(SearchApiTests):
    def setUp(self):
        super().setUp()

        self.api_link = reverse(
            "misago:api:search", kwargs={"search_provider": "users"}
        )



import markdown
from markdown.inlinepatterns import SimpleTagPattern

STRIKETHROUGH_RE = r"(~{2})(.+?)\2"


class StrikethroughExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.registerExtension(self)
        md.inlinePatterns.register(
            SimpleTagPattern(STRIKETHROUGH_RE, "del"), "misago_strikethrough", 100
        )



from django.core.paginator import EmptyPage, InvalidPage
from django.http import Http404

from ...acl.objectacl import add_acl_to_obj
from ...core.cursorpagination import get_page
from ...core.shortcuts import paginate, pagination_dict
from ...threads.permissions import exclude_invisible_threads
from ...threads.serializers import FeedSerializer
from ...threads.utils import add_categories_to_items
from ...threads.viewmodels import ThreadsRootCategory


class UserThreads:
    def __init__(self, request, profile, start=0):
        root_category = ThreadsRootCategory(request)
        threads_categories = [root_category.unwrap()] + root_category.subcategories

        threads_queryset = self.get_threads_queryset(
            request, threads_categories, profile
        )

        posts_queryset = (
            self.get_posts_queryset(request.user, profile, threads_queryset)
            .filter(is_event=False, is_hidden=False, is_unapproved=False)
            .order_by("-id")
        )

        try:
            list_page = get_page(
                posts_queryset, "-id", request.settings.posts_per_page, start
            )
        except (EmptyPage, InvalidPage):
            raise Http404()

        posts = list(list_page.object_list)
        threads = []

        for post in posts:
            threads.append(post.thread)

        add_categories_to_items(
            root_category.unwrap(), threads_categories, posts + threads
        )

        add_acl_to_obj(request.user_acl, threads)
        add_acl_to_obj(request.user_acl, posts)

        self._user = request.user

        self.posts = posts
        self.list_page = list_page

    def get_threads_queryset(self, request, threads_categories, profile):
        return exclude_invisible_threads(
            request.user_acl, threads_categories, profile.thread_set
        )

    def get_posts_queryset(self, user, profile, threads_queryset):
        return profile.post_set.select_related("thread", "poster").filter(
            id__in=threads_queryset.values("first_post_id")
        )

    def get_frontend_context(self):
        return {
            "results": UserFeedSerializer(
                self.posts, many=True, context={"user": self._user}
            ).data,
            "next": self.list_page.next,
        }

    def get_template_context(self):
        return {"posts": self.posts, "next": self.list_page.next}


UserFeedSerializer = FeedSerializer.exclude_fields("poster")



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views.attachments import AttachmentsList, DeleteAttachment
from .views.attachmenttypes import (
    AttachmentTypesList,
    DeleteAttachmentType,
    EditAttachmentType,
    NewAttachmentType,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Attachment
        urlpatterns.namespace("attachments/", "attachments")
        urlpatterns.patterns(
            "attachments",
            path("", AttachmentsList.as_view(), name="index"),
            path("<int:page>/", AttachmentsList.as_view(), name="index"),
            path("delete/<int:pk>/", DeleteAttachment.as_view(), name="delete"),
        )

        # AttachmentType
        urlpatterns.namespace("attachment-types/", "attachment-types", "settings")
        urlpatterns.patterns(
            "settings:attachment-types",
            path("", AttachmentTypesList.as_view(), name="index"),
            path("new/", NewAttachmentType.as_view(), name="new"),
            path("edit/<int:pk>/", EditAttachmentType.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteAttachmentType.as_view(), name="delete"),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Attachments"),
            icon="fas fa-paperclip",
            after="permissions:index",
            namespace="attachments",
        )

        site.add_node(
            name=_("Attachment types"),
            description=_("Specify what files may be uploaded on the forum."),
            parent="settings",
            namespace="attachment-types",
        )



from django.conf import settings

from . import defaults


class StaticSettings:
    def __getattr__(self, name):
        try:
            return getattr(settings, name)
        except AttributeError:
            pass

        try:
            return getattr(defaults, name)
        except AttributeError:
            pass

        raise AttributeError("%s setting is not defined" % name)
